/* Graph query
	Input: A small graph, a big graph, and correspondences.
	Return subgraphs of the big graph.

	Author: Qingwu Yang
	Department of Computer Science
	Texas A&M University, Texas USA
	March, 2006
**/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>

#define TRUE  1
#define FALSE 0

char ** ARGV;
int  ARGC;

//number of outputs
int max_clique = 40;
int top_num = 10;
int *** tops;
int *   tops_size;
double *topscores;

int if_maximum_clique_only = 0;
int if_use_log_score = 1;

double corr_cutoff = 1.0e200;

double edge_gap_penalty = -0.2;  //gap penalty for edges
double mnode_penalty    = -7.0;     //penalty for nodes (neither dummy ones) not in the result

char   queryg_file[50]; // file containing the query graph
char   network_file[50]; // file containing the network 
char   corrfilename[50];
int    max_gap_in_result = 2; 
char   outfname[50];

int if_directed_graph = 0;

//correspondences
typedef struct pair {	
	int v1; //vertex in graph1	
	int v2; //vertex in graph2	
	int id1; //PIN node ID for v1
	int id2; //PIN node ID for v2
	double  sim;      //similarity or other measurement
	double  lgsim;    //logarithm of sim
} pair;

typedef struct vnode {
	int netid;   //id of the node in the network
	int qid;     //id of the corresponding node in the query
	double s;    //similarity score
	double mps;  //maximum possible score
	int no;      //same as coid if not sorted
	int ord;     //total number of neighbors
	int ls;      //number of levels at which the node has edges
	int y;       //
	int *rank;   //number of edges at each level
	struct vnode ***edges; //edges
	double **ewts;//edge weights
} vnode;

typedef struct enode {
	int end1;
	int end2;
	int dist;
	int ord;   //No. XXX edge -- order
	int ls;    //number of levels of edges in which the node has edges
	int tr;    //total number of edges 
	int *rank; //number of edges at each level
	int **edges; //edges
} enode;

//dummy node
typedef struct dummynode {
	int v1; //corresponding vertex in graph1
	int v2; //id in graph2; note! a fake id
	int **edges; //another endpoints
	int n;  //number of edges
	struct dummynode *next;
} dummynode;

//result node
typedef struct rnode {
	double s;   //score
	int    n;   //size
	int ** d;   //data
	struct rnode *prev;
	struct rnode *next;
} rnode;

rnode * Rhead, *Rtail, *Rcurr, *Rtemp;
int curr_max_l;
double curr_max_w;

//cliques finding using DFS
FILE * tmpcf;
int curr_max;
//unsigned long ccnt[400];
int tmpc[30][4];
double tmpcwt[30];
int max_count = 0;

///tree
typedef struct cnode {
	struct tnode * c;
	struct cnode * next;
} cnode;

typedef struct tnode {
	int    qid;
	int    t;     //t=1 the partial path from root to current tnode is a connected subcomponent; 0, otherwise
	double s;
	double x;
	struct tnode *p;
	struct cnode *d;
} tnode;

typedef struct strlist {
	char * s;
	struct strlist * next;
} strlist;

//read in correspondence information
void correspondence_reader(pair ** co, char * fname, int * num, int n, char **names, char **qnames, int q_size);

//calculate scores from sim
void calculate_corrscore(pair * co, int cn);


//create an adjacent matrix graph and allocate space for it
void allocate_space(int *** g, int size);

//given a graph, return number of edges
int count_graph_edge(int ** g, int n);

//given an adjacent matrix containing shortest distances, return number of edges with length less than some cutoff
int count_graph_edge_cutoff(int ** g, int n, int c);

//floyd's algorithm
void floyd(int ** g, int ** dist, int ** path, int size);
//test whether there exists a negative cycle in the graph
char test_negative_cycle(int ** dist, int size);

//read in the distance matrix
void read_dist_matrix(int ** dist, int size, char * fname);
//read in the path matrix
void read_path_matrix(int ** dist, int size, char * fname);
//print the path matrix
void print_path_matrix(int ** dist, int size, char * fname);
//print the distance matrix
void print_dist_matrix(int ** dist, int size, char * fname);
void output_all_pairs_shortest_path(int ** path, int start, int end);

//parse the command-line input
void cmdlin_parser(int argc, char **argv, char *query, char *network, char *corr, char *out, int *e, int *f, int *g, int *h, char *qnfile, char *nnfile);

//info
void info();

//check if one line is blank
int is_blank(char *s);

//read in node names
void nname_reader(char *fname, int * size, char ***nnames);

//read graphs in linked list representation and save to adjacent matrix
void graph_reader_1(char *f, int *** g, int gsize, char ** names);

//put all correspondences together
void corr_extractor(int dn1, pair * co, int cn, int *** conds, int ** con);

//sort the corr graph
void rearrange_corr(int ** conds, int *con, int n);

//create ng -- correspondence nodes as vertices
void ng_builder(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode ***g, int **gsize, int cut, int dcut);

//create ng -- correspondence nodes as vertices
void ng_builder_cc(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode ***g, int **gsize, int cut, int dcut);

//sort the corr wrt similarity
void sort_corr(int ** conds, int *con, pair * co, int cn, int n);

//find cliques in ng -- correspondence nodes as vertices
void clique_finding_ng(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames);

//find cliques with top weights in ng -- correspondence nodes as vertices
void wt_clique_finding_ng(int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, char *outfname);

//create directed ng -- correspondence nodes as vertices
void ng_builder_direct(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode ***g, int **gsize, int cut, int dcut);

//find cliques with top weights in ng -- correspondence nodes as vertices
void wt_clique_finding_ng_direct(int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, int ** conds, int * con, 
	  pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, char * outfname);

////cc
//create linked list version of query
void ll_graph(int ** q, int n, int *** ll, int ** ln);

//build the query tree
void build_query_tree(int **qg, int **qll, int *qlln,int qsize, tnode **qtree, tnode **ntree);

//output the query tree for test
void recurr_output_tree(tnode *tr, int y);

//calculate maximally possible scores in the guide tree
void recurr_score_tree(tnode * tr, int y, int ** dist1, int dn1,int ** dist2, int dn2, vnode **g, int *gsize);

//find connected components with top weights in ng -- correspondence nodes as vertices
void cc_subgraph(tnode * tr, int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, int ** conds, int * con, 
	  pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, char *outfname);


int main(int argc, char ** argv){
	//input graph
	int ** queryg;        // query graph (represented in adjacent matrix)
	int    queryg_size;   // number of nodes in queryg
	int    queryg_edge;   // number of edges in queryg
	char **querygnodenames;
	char   querygnodenames_file[50];
	int ** queryll;
	int *  queryll_n;
	
	//network graph
	//char   network_file[50]; // file containing the network 
	int ** network;       // queried graph (represented in adjacent matrix)
	int    network_size;  // number of nodes in network
	int    network_edge;  // number of edges in network
	char   network_shrt_dist_fname[50], network_shrt_path_fname[50];
	int ** network_shrt_dist;
	int ** network_shrt_path;
	char **networknodenames;
	char   networknodenames_file[50];

	//input correspondences
	//char corrfilename[50];
	pair * corr;          // hold the correspondence information
	int    corr_size;     // number of correspondences

	//preprocessing
	int prepro_edge_cutoff = 1;
	int if_apply_shrt_alg_on_queryg  = 1;
	int if_apply_shrt_alg_on_network = 1;

	int if_results_cc = 1;
	//int  max_gap_in_result = 2; 

	//char outfname[50];

	//graph with nodes as nodes
	int ** conodes;
	int *  conum;

	vnode ** ng;
	int  * ng_size;

	//guide tree
	tnode * qtree; //nodes are in increasing order on each path
	tnode * ntree; //nodes are naturally ordered

	ARGV = argv;
	ARGC = argc;
	
	//system("date");
	strcpy(queryg_file,   "query");
	strcpy(network_file,  "input");
	strcpy(corrfilename,   "corr");
	strcpy(outfname, "result");
	strcpy(querygnodenames_file, "querynames");
	strcpy(networknodenames_file, "networknames");

	//parse the command-line input
	cmdlin_parser(argc, argv, queryg_file, network_file, corrfilename, outfname,&max_gap_in_result, &if_apply_shrt_alg_on_queryg,  
		           &if_apply_shrt_alg_on_network, &if_results_cc, querygnodenames_file, networknodenames_file);				  

	strcpy(network_shrt_dist_fname, network_file);
	strcat(network_shrt_dist_fname, "_shrtdist");
	strcpy(network_shrt_path_fname, network_file);
	strcat(network_shrt_path_fname, "_shrtpath");

	//read query node names and size
	nname_reader(queryg_file, &queryg_size, &querygnodenames);
	//read query as a graph
	graph_reader_1(queryg_file, &queryg, queryg_size, querygnodenames);
	//calculate number of edges
	queryg_edge = count_graph_edge(queryg, queryg_size); 

	//read in node names of the network
	nname_reader(network_file, &network_size, &networknodenames);
	//read network as a graph
	graph_reader_1(network_file, &network, network_size, networknodenames);
	//calculate number of edges
	network_edge = count_graph_edge(network, network_size);

	//read in the correspondence information
	correspondence_reader(&corr, corrfilename,  &corr_size, network_size, networknodenames, querygnodenames, queryg_size);

	//create dist and path graphs
	allocate_space(&network_shrt_dist, network_size);
	allocate_space(&network_shrt_path, network_size);

	//get shortest path for network
	if( if_apply_shrt_alg_on_network ){
		//floyd's algorithm to calculate the all-pairs shortest paths
		floyd(network, network_shrt_dist, network_shrt_path, network_size);

		//test whether there exists a negative cycle in the graph
		if( test_negative_cycle(network_shrt_dist, network_size) ){
			fprintf(stdout, "There is a negative cycle in the graph -- %s!\n", network_file);
			fflush(NULL), exit(0);
		}

		//print_dist_matrix(network_shrt_dist, network_size, network_shrt_dist_fname);
		//print_path_matrix(network_shrt_path, network_size, network_shrt_path_fname);
	} else {
		//read in directly
		read_dist_matrix(network_shrt_dist, network_size, network_shrt_dist_fname);
		read_path_matrix(network_shrt_path, network_size, network_shrt_path_fname);
	}

	//main calculation
	if( if_results_cc ){
		//create linked list version of query
		ll_graph(queryg, queryg_size, &queryll, &queryll_n);

		//build the query tree
		build_query_tree(queryg, queryll, queryll_n, queryg_size, &qtree, &ntree);

 		//fprintf(stdout, "\n\nGuide tree built\n"), fflush(NULL);
		//system("date");

		//recurr_output_tree(qtree, 0);
		//recurr_output_tree(ntree, 0);

		max_clique = queryg_size+1;

		//calculate scores from sim
		calculate_corrscore(corr, corr_size);

		//put all correspondences together
		corr_extractor(queryg_size, corr, corr_size, &conodes, &conum);

		//sort the corr wrt similarity
		sort_corr(conodes, conum, corr, corr_size, queryg_size);

		if( if_directed_graph == 0 ){
			//create ng -- correspondence nodes as vertices
			ng_builder_cc(queryg, queryg_size, network_shrt_dist, network_size, conodes, conum, 
						corr, corr_size, &ng, &ng_size, prepro_edge_cutoff, max_gap_in_result);

			//calculate maximally possible scores in the guide tree
			recurr_score_tree(ntree, 0, queryg, queryg_size,network_shrt_dist, network_size, ng, ng_size);

			//find connected components with top weights in ng -- correspondence nodes as vertices
			cc_subgraph(ntree, queryg, queryg_size, network_shrt_dist, network_shrt_path, network_size, conodes, conum, 
					corr, corr_size, ng, ng_size, prepro_edge_cutoff, max_gap_in_result, querygnodenames, networknodenames, outfname);
		} else {
			//create directed ng -- correspondence nodes as vertices
			ng_builder_cc(queryg, queryg_size, network_shrt_dist, network_size, conodes, conum, 
						corr, corr_size, &ng, &ng_size, prepro_edge_cutoff, max_gap_in_result);

			//calculate maximally possible scores in the guide tree
			recurr_score_tree(ntree, 0, queryg, queryg_size,network_shrt_dist, network_size, ng, ng_size);

			//find connected components with top weights in ng -- correspondence nodes as vertices
			cc_subgraph(ntree, queryg, queryg_size, network_shrt_dist, network_shrt_path, network_size, conodes, conum, 
					corr, corr_size, ng, ng_size, prepro_edge_cutoff, max_gap_in_result, querygnodenames, networknodenames, outfname);
		}
	} else {
		max_clique = queryg_size+1;

		//calculate scores from sim
		calculate_corrscore(corr, corr_size);

		//put all correspondences together
		corr_extractor(queryg_size, corr, corr_size, &conodes, &conum);

		//sort the corr wrt similarity
		sort_corr(conodes, conum, corr, corr_size, queryg_size);

		if( if_directed_graph == 0 ){
			//create ng -- correspondence nodes as vertices
			ng_builder(queryg, queryg_size, network_shrt_dist, network_size, conodes, conum, 
						corr, corr_size, &ng, &ng_size, prepro_edge_cutoff, max_gap_in_result);

			//find cliques with top weights in ng -- correspondence nodes as vertices
			wt_clique_finding_ng(queryg, queryg_size, network_shrt_dist, network_shrt_path, network_size, conodes, conum, 
					corr, corr_size, ng, ng_size, prepro_edge_cutoff, max_gap_in_result, querygnodenames, networknodenames, outfname);
		} else {
			//create directed ng -- correspondence nodes as vertices
			ng_builder_direct(queryg, queryg_size, network_shrt_dist, network_size, conodes, conum, 
						corr, corr_size, &ng, &ng_size, prepro_edge_cutoff, max_gap_in_result);

			//find cliques with top weights in ng -- correspondence nodes as vertices
			wt_clique_finding_ng_direct(queryg, queryg_size, network_shrt_dist, network_shrt_path, network_size, conodes, conum, 
					corr, corr_size, ng, ng_size, prepro_edge_cutoff, max_gap_in_result, querygnodenames, networknodenames, outfname);
		}
	}
				
/*
	fprintf(stdout, "\n\n"), fflush(NULL);
	system("date");

	fprintf(stdout, "\n\nParameters:");
	fprintf(stdout, "\ncorr_cutoff = %7.2f", corr_cutoff);
	fprintf(stdout, "\nif_maximum_clique_only = %d", if_maximum_clique_only);
	fprintf(stdout, "\nmax_gap_in_result = %d", max_gap_in_result);
	fprintf(stdout, "\nif_results_cc = %d", if_results_cc);
	fprintf(stdout, "\nif_apply_shrt_alg_on_queryg = %d", if_apply_shrt_alg_on_queryg);
	fprintf(stdout, "\nif_apply_shrt_alg_on_network = %d", if_apply_shrt_alg_on_network);
	fprintf(stdout, "\nedge_gap_penalty = %7.2f \nmnode_penalty = %7.2f\n", 
		            edge_gap_penalty, mnode_penalty);
	
	fprintf(stdout, "\n\n***DONE***"), fflush(NULL);
*/
	return 0;
} //main()



//read in node names
void nname_reader(char *fname, int * size, char ***nnames){
	FILE * inf;
	char *buff, *buffer, ** s;
	int n, max;
	char tmpstr[200], ch;
	char *ptr;

	strlist * tail, *curr;
	strlist * slist = NULL;

	if( !(inf = fopen(fname, "r")) ){
		*nnames = NULL;
		return;
	} 

	//determine the max line length
	max = 0; 
	n = 0;
	while( (ch = fgetc(inf)) != EOF ){
		if( ch == '\n' ){
			++n;
			if( n > max )
				max = n;
			n  = 0;
		} else 
			++ n;		
	}
	if( n > max )
		max = n;
	max += 2;

	if( !(buff = malloc(max * sizeof(char))) )
		fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);
	if( !(buffer = malloc(max * sizeof(char))) )
		fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

	rewind(inf);

	//read in the first column
	while( fgets(buff, max, inf) ){
		if( ! is_blank( buff ) ){
			if( ! slist ){
				if( !( slist = calloc(1, sizeof(strlist))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				if( sscanf(buff, " %s %*s ", tmpstr) != 1 )
					fprintf(stderr, "\nwrong here\n%s", buff), fflush(NULL), exit(0);

				if( !( slist->s = calloc(strlen(tmpstr)+1, sizeof(char))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				strcpy(slist->s, tmpstr);
				slist->next = NULL;

				tail = slist;
			} else {
				if( !( curr = calloc(1, sizeof(strlist))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				if( sscanf(buff, " %s %*s ", tmpstr) != 1 )
					fprintf(stderr, "\nwrong here\n%s", buff), fflush(NULL), exit(0);

				if( !( curr->s = calloc(strlen(tmpstr)+1, sizeof(char))) )
					fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

				strcpy(curr->s, tmpstr);
				curr->next = NULL;

				tail->next = curr;
				tail = tail->next;
			}			
		}
	}

	rewind(inf);

	//read other names, if any
	while( fgets(buff, max, inf) ){
		if( ! is_blank( buff ) ){
			strcpy(buffer, buff);

			ptr = strtok(buffer, ", \n");

			while( (ptr=strtok(NULL, ", \n")) ) {				
				strcpy(tmpstr, ptr);

				curr = slist;
				while( curr ){
					if( ! strcmp(curr->s, tmpstr) ){
						break;
					}

					curr = curr->next;
				}

				if( ! curr ){
					if( !( curr = calloc(1, sizeof(strlist))) )
						fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

					if( !( curr->s = calloc(strlen(tmpstr)+1, sizeof(char))) )
						fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

					strcpy(curr->s, tmpstr);
					curr->next = NULL;

					tail->next = curr;
					tail = tail->next;
				}				
			}			
		}
	}

	//count # of nodes
	n = 0;
	curr = slist;
	while( curr ){
		++ n;
		curr = curr->next;
	}

	*size = n;

	if( !(s = calloc(n, sizeof(char *))) )
		fprintf(stderr, "\nout of memory!\n"), fflush(NULL), exit(0);

	n = 0;
	curr = slist;
	while( curr ){
		if( !(s[n] = calloc(strlen(curr->s)+1, sizeof(char))) )
			fprintf(stderr, "\nout of memory!\n"), fflush(NULL), exit(0);

		strcpy(s[n], curr->s);
		
		++ n;
		curr = curr->next;
	}
	
	fclose(inf);

	*nnames = s;

	return;
}


//read graphs in linked list representation and save to adjacent matrix
void graph_reader_1(char *f, int *** g, int gsize, char ** names){
	FILE * inf;
	int ** h;
	int  i, j;//hsize,
	char *buffer;
	char *token;
	int max, n;
	char ch;

	if( !(inf = fopen(f, "r")) )
		fprintf(stdout, "File %s can't be opened for reading!\n", f), fflush(NULL), exit(0);

	//allocate space for the graph
	if( !(h = (int **)calloc(gsize, sizeof(int *))) )
		fprintf(stdout, "out of memory for h -- %s\n", f), fflush(NULL), exit(0);
	for(i=0; i<gsize; i++)
		if( !(h[i] = (int *)calloc(gsize, sizeof(int))) )
			fprintf(stdout, "out of memory for h[i] -- %s\n", f), fflush(NULL), exit(0);

	for(i=0; i<gsize; i++){
		for(j=0;j<gsize; j++)
			h[i][j] = SHRT_MAX;
	}

	//determine the max line length
	max = 0; 
	n = 0;
	while( (ch = fgetc(inf)) != EOF ){
		if( ch == '\n' ){
			++n;
			if( n > max )
				max = n;
			n  = 0;
		} else 
			++ n;		
	}
	if( n > max )
		max = n;
	max += 2;

	if( !(buffer = malloc(max * sizeof(char))) )
		fprintf(stderr, "\nout of memory\n"), fflush(NULL), exit(0);

	rewind(inf);

	while( fgets(buffer, max, inf) ){
		if( !is_blank(buffer) ){
			if( (token=strtok(buffer, ", \n")) ) {
				for(i=0; i<gsize; ++i){
					if( ! strcmp(names[i], token) )
						break;
				}

				if( i == gsize )
					fprintf(stdout, "\ni == gsize\n"), fflush(NULL), exit(0);				
			}

			while( (token=strtok(NULL, ", \n")) ) {
				for(j=0; j<gsize; ++j){
					if( ! strcmp(names[j], token) )
						break;
				}

				if( j == gsize ){
					fprintf(stdout, "\n%s\n", token);
					fprintf(stdout, "\nj == gsize\n"), fflush(NULL), exit(0);	
				}

				if( j >= 0 )
					h[i][j] = 1;
			}	
		}
	}

	*g = h;

	fclose( inf );

	for(i=0; i<gsize; i++){
		for(j=i+1;j<gsize; j++){
			if( h[i][j] != h[j][i] ){
				if_directed_graph = 1;
				return;
			}
		}
	}

	return;
}



void dfs_clique_max_only(enode ** g, int * gsize, int n, int l){
	int j,k,x,m, con=0;//, prune=0;

	if( g[ tmpc[l-1][0] ][ tmpc[l-1][1] ].ls > 0 ){
		//for(i=0; i< (n-tmpc[l-1][0]-2); i++){ //pruning
			if( (g[ tmpc[l-1][0] ][ tmpc[l-1][1] ].rank)[0] > 0 ){
				x= tmpc[l-1][0] +1;
				for(j=1; j< gsize[x]; j++){
					if( g[x][j].ls >= curr_max-l-1 ){ //pruning  (((curr_max-l)>0)?(curr_max-l):0) 
						//check if connected with enodes already in tmpc
						for(k=0; k<l; k++){
							if( g[ tmpc[k][0] ][0].end1 == g[x][0].end1 ){
								if( g[ tmpc[k][0] ][ tmpc[k][1] ].end1 != g[x][j].end1 )
									break;						
							} else if( g[ tmpc[k][0] ][0].end1 == g[x][0].end2 ){
								if( g[ tmpc[k][0] ][ tmpc[k][1] ].end1 != g[x][j].end2 )
									break;						
							} else if( g[ tmpc[k][0] ][0].end2 == g[x][0].end1 ){
								if( g[ tmpc[k][0] ][ tmpc[k][1] ].end2 != g[x][j].end1 )
									break;						
							} else if( g[ tmpc[k][0] ][0].end2 == g[x][0].end2 ){
								if( g[ tmpc[k][0] ][ tmpc[k][1] ].end2 != g[x][j].end2 )
									break;						
							} else {										
							} 	
						}

						if( k == l ){
							con = 1;
							tmpc[l][0] = x;
							tmpc[l][1] = j;
							dfs_clique_max_only(g, gsize, n, l+1);
						}
					} //if( g[x][j].ls >= curr_max - l)
				}
			} 
		//}
	} 

	if( con == 0 && l >= curr_max ){
		if( !(tmpcf = fopen("tempc", "a")) )
			fprintf(stdout, "FILE tempc can't be opened for writing!\n"), fflush(NULL), exit(0);

		fprintf(tmpcf, "\ntop%02d -- cliqu%02d: ", max_count+1, curr_max);

		for( m=0; m<l; m++){
			fprintf(tmpcf, "\n\t(%5d, %5d, %2d, %2d, %4d, %4d)", g[tmpc[m][0]][tmpc[m][1]].end1, 
					g[tmpc[m][0]][tmpc[m][1]].end2, 
					g[tmpc[m][0]][tmpc[m][1]].dist, g[tmpc[m][0]][tmpc[m][1]].ord, 
					g[tmpc[m][0]][0].end1 +1, g[tmpc[m][0]][0].end2+1);
		}

		max_count ++;

		fclose(tmpcf);
	}

	return;
}

//given an adjacent matrix containing shortest distances, return number of edges with length less than some cutoff
int count_graph_edge_cutoff(int ** g, int n, int c){
	int i,j, m=0;

	for(i=0; i<n; i++)
		for(j=i+1; j<n; j++)
			if( g[i][j] <= c )
				m++;

	return m;
}


//given a graph, return number of edges
int count_graph_edge(int ** g, int n){
	int i,j, m=0;

	for(i=0; i<n; i++)
		for(j=i+1; j<n; j++)
			if( g[i][j] == 1 )
				m++;

	return m;
}

//print the distance matrix
void print_dist_matrix(int ** dist, int size, char * fname){
	FILE * outf;
	int i, j;
	
	if( !(outf=fopen(fname, "w")) )
		fprintf(stdout, "%s cannot be opened for writing!\n", fname), exit(0);

	for(i=0;i<size;i++){
		fprintf(outf, "\n %1d", i);
		for(j=0;j<size;j++){
			fprintf(outf, " %1d", dist[i][j]);
		}
	}

	fclose(outf);

	return;
}

//print the path matrix
void print_path_matrix(int ** dist, int size, char * fname){
	FILE * outf;
	int i, j;
	
	if( !(outf=fopen(fname, "w")) )
		fprintf(stdout, "%s cannot be opened for writing!\n", fname), exit(0);

	for(i=0;i<size;i++){
		fprintf(outf, "\n %1d", i);
		for(j=0;j<size;j++){
			fprintf(outf, " %1d", dist[i][j]);
		}
	}
	//fprintf(stdout, "\n\n");
	fclose(outf);

	return;
}

//read in the path matrix
void read_path_matrix(int ** dist, int size, char * fname){
	FILE * inf;
	int i, j;
	long len;
	char * buff;
	char * token;

	len=(size+1)*10;

	if( !(buff=malloc(len*sizeof(char))) )
		fprintf(stdout, "out of memory!\n"), exit(0);
	
	if( !(inf=fopen(fname, "r")) )
		fprintf(stdout, "%s cannot be opened for reading!\n", fname), exit(0);

	i=0;
	while( fgets(buff, len, inf) ){
		if(strlen(buff) > 20){
			if( (i+1) > size)
				fprintf(stdout, "(more) data error in the file %s\n", fname), exit(0); 

			token=strtok(buff, " ");
			i=atoi(token);
			j=0;
			while( (token=strtok(NULL, " ")) ){
				if( (j+1) > size)
					fprintf(stdout, "(more) data error in the line %d (in the file %s)\n", i, fname), exit(0); 
				dist[i][j++]=atoi(token);
			}
			if(j != size)
				fprintf(stdout, "(less) data error in the line %d (in the file %s)\n", i, fname), exit(0); 
		}
	}

	if( (i+1) != size)
		fprintf(stdout, "(less) data error in the file %s\n", fname), exit(0); 

	fclose(inf);
	free(buff);

	return;
}

//read in the distance matrix
void read_dist_matrix(int ** dist, int size, char * fname){
	FILE * inf;
	int i, j;
	long len;
	char * buff;
	char * token;

	len=(size+1)*10;

	if( !(buff=malloc(len*sizeof(char))) )
		fprintf(stdout, "out of memory!\n"), exit(0);
	
	if( !(inf=fopen(fname, "r")) )
		fprintf(stdout, "%s cannot be opened for reading!\n", fname), exit(0);

	i=0;
	while( fgets(buff, len, inf) ){
		if(strlen(buff) > 20){
			if( (i+1) > size)
				fprintf(stdout, "(more) data error in the file %s\n", fname), exit(0); 

			token=strtok(buff, " ");
			i=atoi(token);
			j=0;
			while( (token=strtok(NULL, " ")) ){
				if( (j+1) > size)
					fprintf(stdout, "(more) data error in the line %d (in the file %s)\n", i, fname), exit(0); 
				dist[i][j++]=atoi(token);
			}
			if(j != size)
				fprintf(stdout, "(less) data error in the line %d (in the file %s)\n", i, fname), exit(0); 
		}
	}

	if( (i+1) != size)
		fprintf(stdout, "(less) data error in the file %s\n", fname), exit(0); 

	fclose(inf);
	free(buff);
	return;
}

//hold the shortest path recovered
int spath[100];
int spath_count;

//recover the shortest path 
void output_all_pairs_shortest_path(int ** path, int start, int end) {
//fprintf(stdout, "oapsp:  start=%d end=%d  %d\n", start, end, path[start][end]);
	if( start == end)  {
		//fprintf(stdout, " %2d", start);
		spath[spath_count++]=start;
	} else if (path[start][end] == -1){
		//fprintf(stdout, " No path from %2d to %2d!\n", start, end);
		spath[spath_count]=-1;
	} else {
		output_all_pairs_shortest_path(path, start, path[start][end]);
		//fprintf(stdout, " %2d", end);
		spath[spath_count++]=end;
	}
}


//floyd's algorithm
void floyd(int ** g, int ** dist, int ** path, int size){
	int i, j, k;

	//Initialization
	for(i=0; i<size;i++){
		for(j=0;j<size;j++){
			dist[i][j]=g[i][j];
			if( g[i][j] >= SHRT_MAX ){
				path[i][j]=-1;
			} else 
				path[i][j]=i;
		}
		//dist[i][i]=0;	
	}

	//The working loop 
	for(k=0;k<size;k++){
		for(i=0;i<size;i++)
			for(j=0;j<size;j++)
				if( (dist[i][k] != SHRT_MAX) && (dist[k][j] != SHRT_MAX) 
					&& (dist[i][k]+dist[k][j] < dist[i][j])  ){
					dist[i][j]=dist[i][k]+dist[k][j];
					path[i][j]=path[k][j];
				}
	}

	return;
}

//test whether there exists a negative cycle in the graph
char test_negative_cycle(int ** dist, int size){
	int i;

	for(i=0;i<size;i++)
		if(dist[i][i]<0)
			return TRUE;

	return FALSE;
}

//create an adjacent matrix graph and allocate space for it
void allocate_space(int *** g, int size){
	int i;

	if( !( *g = (int **)malloc(size * sizeof(int *))) )
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);
	for(i=0;i<size;i++)
		if( !( (*g)[i]= malloc(size * sizeof(int))) )
			fprintf(stdout, "out of memory!\n"), exit(0);
}

//check if one line is blank
int is_blank(char *s){
	while( *s != '\n' && *s != '\0' && *s != EOF ){
		if( isalnum( *s ) )
			return 0;
		s++;
	}

	return 1;
}


//read in correspondence information
void correspondence_reader(pair ** corr, char * fname, int * num, int n, char **names, char **qnames, int q_size){
	FILE * inf;
	char buffer[4096], str[200], str2[200];
	int i, j, k;
	double d;
	pair * co;

	if( !(inf=fopen(fname, "r")) )
		fprintf(stdout, "The file %s cannot be opened for reading!\n", fname), exit(0);

	i=0;
	while( fgets(buffer, 4096, inf) ) {
		if( ! is_blank(buffer) )
			++ i;
	}
	*num = i;
	if( !( *corr = co = calloc(i, sizeof(pair))) )
		fprintf(stdout, "out of memory!\n"), fflush(NULL), exit(0);

	rewind(inf);

	k = 0;
	while( fgets(buffer, 4096, inf) ){
		if( ! is_blank(buffer) ){		
			if (sscanf(buffer, " %s %s %le ", str, str2, &d) == 3){
				for( i=0; i<q_size; ++i ){
					if( ! strcmp(qnames[i], str) )
						break;
				}
				if( i == q_size )
					fprintf(stdout, "\ncorr: i == q_size\n"), fflush(NULL), exit(0);

				for( j=0; j<n; ++j ){
					if( ! strcmp(names[j], str2) )
						break;
				}
				if( j == n ){
					fprintf(stdout, "\nstr2: %s\n", str2);
					fprintf(stdout, "\ncorr: j == n\n"), fflush(NULL), exit(0);
				}

				co[k].v1 = i;
				co[k].v2 = j;
				co[k].sim = d;
			}  else {
				fprintf(stdout, "*3* Missing data in the file %s.\n", fname);
				fflush(NULL), exit(0);
			}
			
			k++;
		}
	}
	fclose(inf);
	return;
}


//calculate scores from sim
void calculate_corrscore(pair * co, int cn){
	int i;
	double s;

	s = - log( 1e-200 );
	for(i=0; i<cn; i++){
		if( if_use_log_score == 1 ){
			if( co[i].sim >= 0 ){
				if( co[i].sim <= 1e-200 ){
					co[i].lgsim = s;
				} else {
					co[i].lgsim = - log( co[i].sim );
				}
			} else {
				fprintf(stdout, "\n\nNegative Similarity Score!\n\n"), fflush(NULL), exit(0);
			}
		} else {
			co[i].lgsim = - co[i].sim;
		}
	}
}

//	cmdlin_parser(argc, argv, queryg_file, network_file, corrfilename, outfname,&max_gap_in_result,
//		&if_apply_shrt_alg_on_queryg,  
//        &if_apply_shrt_alg_on_network, &if_results_cc, querygnodenames_file, networknodenames_file);	

//parse the command-line input
void cmdlin_parser(int argc, char **argv, char *query, char *network, char *corr, char *out, 
				   int *e, int *f, int *g, int *h, char *qnfile, char *nnfile){
	int i;

	if( argc < 4 ){
		info();
	} else {
		for(i=1; i<argc; i++){
			if( strstr(argv[i], "-q") ){
				if( sscanf(argv[i], "-q=%s", query) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-i") ){
				if( sscanf(argv[i], "-i=%s", network) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-c") ){
				if( sscanf(argv[i], "-c=%s", corr) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-b") ){
				if( sscanf(argv[i], "-b=%s", qnfile) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-j") ){
				if( sscanf(argv[i], "-j=%s", nnfile) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-g") ){
				if( sscanf(argv[i], "-g=%d", e) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();

				++ (*e);
			} else if( strstr(argv[i], "-f") ){
				if( sscanf(argv[i], "-f=%d", f) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-e") ){
				if( sscanf(argv[i], "-e=%d", g) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-h") ){
				if( sscanf(argv[i], "-h=%d", h) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-n") ){
				if( sscanf(argv[i], "-n=%d", &top_num) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-l") ){
				if( sscanf(argv[i], "-l=%d", &if_use_log_score) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-o") ){
				if( sscanf(argv[i], "-o=%s", out) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			} else if( strstr(argv[i], "-s=") ){
				if( sscanf(argv[i], "-s=%lf", &corr_cutoff) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();			
			} else if( strstr(argv[i], "-v=") ){
				if( sscanf(argv[i], "-v=%lf", &mnode_penalty) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();	
				mnode_penalty *= (-1);
			} else if( strstr(argv[i], "-p=") ){
				if( sscanf(argv[i], "-p=%lf", &edge_gap_penalty) != 1 )
					fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();		
				edge_gap_penalty *= (-1);
			} else {
				fprintf(stdout, "\nWrong command line input: %s\n", argv[i]), fflush(NULL), info();
			}
		}
	}
}

//info
void info(){
	fprintf(stdout, "\ngraphmatch \n\t-q=\"file name containing query graph\" "
					"\n\t-i=\"file name containing input graph\""
					"\n\t-c=\"file name containing correspondences between vertices\""
	                "\n\t-g=maximum number of indels between two matches that are adjacent in the query graph"
					"\n\t-l= \"type of match score\""
					"\n\t\t 0 -- use similarity score as match score"
					"\n\t\t 1 -- use negative logarithm of similarity score as match score"
					"\n\t-p=\"indel penalty\""
					"\n\t-v=\"non-associated vertex penalty\""		
					"\n\t-n=\"number of output graphs\""
					"\n\t-o=\"output file name\""
					"\n\n");
					//"\n\t-j=\"network node name file\""
					//"\n\t-f=apply_shrt_alg_on_query(0 or 1) -g=apply_shrt_alg_on_network(0 or 1)"
					//"\n\t-s=similarity_cutoff"	                
					//"\n\t-h=if_results_cc -o=output_file -t=top#output\n");
	fflush(NULL), exit(0);
}


//put all correspondences together
void corr_extractor(int dn1, pair * co, int cn, int *** conds, int ** con){
	int i,j, p;

	int  ** coids;
	int  *  coids_n;

	if( !(coids = (int **)calloc(dn1, sizeof(int*))) )
		fprintf(stdout, "out of memory for coids in eg_builder_dummy()\n"), fflush(NULL), exit(0);
	if( !(coids_n = (int *)calloc(dn1, sizeof(int))) )
		fprintf(stdout, "out of memory for coids_n in eg_builder_dummy()\n"), fflush(NULL), exit(0);

	//printf("\n*** Correspondences ***");
	//put all corrs together
	for(i=0; i<dn1; i++){
		coids_n[i] = 0;

		for(p=0; p<cn; p++)
			if( co[p].v1 == i && co[p].sim <= corr_cutoff )
				coids_n[i] ++;
		coids_n[i]++;
		
		if( !(coids[i] = (int *)malloc(coids_n[i] * sizeof(int))) )
			fprintf(stdout, "out of memory for coids[] in eg_builder_dummy()\n"), fflush(NULL), exit(0);
		
		//printf("\n%2d: (%3d)", i+1, coids_n[i]-1);
		coids[i][0] = i;
		for(j=1, p=0; p<cn; p++){
			if( co[p].v1 == i && co[p].sim <= corr_cutoff ){
				//printf(" %d", co[p].v2);
				coids[i][j++] = co[p].v2;							
			}			
		}		
	}
	//printf("\n");

	*conds = coids;
	*con   = coids_n;

	return;
}


//sort the corr wrt similarity
void sort_corr(int ** conds, int *con, pair * co, int cn, int n){
	int i,j, p, q, m;
	double * ts, min;

	q = 0;
	for(i=0; i<n; i++){
		if( con[i] > q )
			q = con[i];
	}
	if( !(ts = calloc(q, sizeof(double))) )
		fprintf(stdout, "out of memory for ti in sort_corr()\n"), fflush(NULL), exit(0);

	for(i=0; i<n; i++){
		ts[0] = 0;
		for(j=1; j<con[i]; j++){
			for(p=0; p<cn; p++){
				if( co[p].v1 == i && co[p].v2 == conds[i][j] ){
					ts[j] = co[p].sim;
					break;
				}
			}
		}

		//sort the nonzero numbers		
		for(j=1; j<con[i]; j++){
			min = ts[j];
			m   = j;
			for(p=j+1; p<con[i]; p++){
				if( ts[p] < min ){
					min = ts[p];
					m   = p;
				}
			}

			if( min < ts[j] ){
				ts[m] = ts[j];
				ts[j] = min;

				q = conds[i][j];
				conds[i][j] = conds[i][m];
				conds[i][m] = q;
			}
		}
	}

	if( ts )
		free(ts);
	return;
}



//sort the corr graph
void rearrange_corr(int ** conds, int *con, int n){
	int i,j, k, max;
	int * ptr;

	for(i=0; i<n; i++){
		max = 0;
		for(j=i+1; j<n; j++){
			if( con[j] > max ){
				max = con[j];
				k = j;
			}
		}

		if( con[i] < max ){
			con[k] = con[i];
			con[i] = max;
			
			ptr     = conds[k];
			conds[k]= conds[i];
			conds[i]= ptr;
		}
	}

	return;
}

//double tmpwwtt;
//=============================================================================================
//output top cliques
void output_top_cliques_ng(pair * co, int cn, int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, 
	                       int cut, int dcut, char **qnames, char **nnames, char *outfname){
	int i, j, k, m, n;
	double low = 1.0e200, high=-1.0e200;

	if( !(tmpcf = fopen(outfname, "w")) )
		fprintf(stdout, "FILE tempc can't be opened for writing!\n"), fflush(NULL), exit(0);

	Rcurr = Rhead;
	n = 9999;
	m = 0;
	while( Rcurr && Rcurr->n >0 ){
		if(Rcurr->n < n)
			n = Rcurr->n;

		if(Rcurr->n > m)
			m = Rcurr->n;

		if(Rcurr->s < low)
			low = Rcurr->s;

		if(Rcurr->s > high)
			high = Rcurr->s;

		Rcurr = Rcurr->next;
	}

	//output parameters
	fprintf(tmpcf, "%s ", ARGV[0]);
	fprintf(tmpcf, "-q=%1s ", queryg_file);
	fprintf(tmpcf, "-i=%1s ", network_file);
	fprintf(tmpcf, "-c=%1s ", corrfilename);
	fprintf(tmpcf, "-g=%1d ", max_gap_in_result -1);
	fprintf(tmpcf, "-l=%1d ", if_use_log_score);
	fprintf(tmpcf, "-p=%3.1f ", - edge_gap_penalty);
	fprintf(tmpcf, "-v=%3.1f ", - mnode_penalty);
	fprintf(tmpcf, "-n=%1d ", top_num);
	fprintf(tmpcf, "-o=%1s ", outfname);
	//fprintf(tmpcf, "\n");

	//for(i=0; i<ARGC; ++i)
	//	fprintf(tmpcf, "%s ", ARGV[i]);
	//fprintf(tmpcf, "\n");
	//fprintf(tmpcf, "\n\n\nGap penalty:  %7.2f \nMiss node penalty:  %7.2f\n", edge_gap_penalty, mnode_penalty);
	//fprintf(tmpcf, "\nSize of query: %d \nmax_size = %d\nmin_size = %d\nmax_score = %7.2f\nmin_score = %7.2f\n", dn1, m, n, high, low);

	//fprintf(tmpcf, "\ntmpwwtt = %7.2f\n", tmpwwtt);

	Rcurr = Rhead;
	i = 1;
	while( Rcurr && Rcurr->n >0 ){
		//fprintf(tmpcf, "\n\nResult%02d: SIZE %2d, SCORE %7.2f ", i, Rcurr->n, Rcurr->s);
		//fprintf(tmpcf, "\n\nResult%02d: ( SCORE= %5.2f ) ", i, Rcurr->s);
		fprintf(tmpcf, "\n\nResult%02d: score=%3.2f  ", i, Rcurr->s);
		fprintf(tmpcf, "\nmatches:");
		for(j=0; j<Rcurr->n; j++){
			if( qnames && qnames[Rcurr->d[j][0]] ){
				fprintf(tmpcf, "\n\t %-10s ", qnames[Rcurr->d[j][0]]);
			} else {
				fprintf(tmpcf, "\n\t %-10d ", Rcurr->d[j][0]+1);
			}

			if( nnames && nnames[Rcurr->d[j][1]] ){
				fprintf(tmpcf, " %-10s ", nnames[Rcurr->d[j][1]]);
			} else {
				fprintf(tmpcf, " %-10d ", Rcurr->d[j][1]+1);
			}


			//fprintf(tmpcf, " %5d (%2d),", Rcurr->d[j][1], Rcurr->d[j][0] +1);				
		}

		fprintf(tmpcf, "\nedges:");
		if( if_directed_graph == 0 ){
			for(m=0; m<Rcurr->n; m++){
				for(n=m+1; n<Rcurr->n; n++){
					if( dist1[ Rcurr->d[m][0] ][ Rcurr->d[n][0] ] <= cut ){

						if( qnames && qnames[Rcurr->d[m][0]] ){
							fprintf(tmpcf, "\n\t %-10s", qnames[Rcurr->d[m][0]]);
						} else {
							fprintf(tmpcf, "\n\t %-10d", Rcurr->d[m][0]+1);
						}

						if( qnames && qnames[Rcurr->d[n][0]] ){
							fprintf(tmpcf, " -- %-10s ", qnames[Rcurr->d[n][0]]);
						} else {
							fprintf(tmpcf, " -- %-10d ", Rcurr->d[n][0]+1);
						}

						if( dist2[ Rcurr->d[m][1] ][ Rcurr->d[n][1] ] > 1 ){
							spath_count=0;
							output_all_pairs_shortest_path(path2, Rcurr->d[m][1], Rcurr->d[n][1]);

							if( nnames && nnames[ spath[0] ] ){
								fprintf(tmpcf, "  %1s", nnames[ spath[0] ]);
							} else {
								fprintf(tmpcf, "  %1d",  spath[0] +1);
							}
							
							for(k=1; k<spath_count; k++){
								if( nnames && nnames[ spath[k] ] ){
									fprintf(tmpcf, " -- %1s", nnames[ spath[k] ]);
								} else {
									fprintf(tmpcf, " -- %1d",  spath[k] +1);
								}
							}
						} else {
							if( nnames && nnames[ Rcurr->d[m][1] ] ){
								fprintf(tmpcf, "  %1s", nnames[ Rcurr->d[m][1] ]);
							} else {
								fprintf(tmpcf, "  %1d",  Rcurr->d[m][1] +1);
							}

							if( nnames && nnames[ Rcurr->d[n][1] ] ){
								fprintf(tmpcf, " -- %1s", nnames[ Rcurr->d[n][1] ]);
							} else {
								fprintf(tmpcf, " -- %1d",  Rcurr->d[n][1] +1);
							}						
						}

						//fprintf(tmpcf, "\n\t%5d -- %5d (%2d -- %2d) (%d)", Rcurr->d[m][1], Rcurr->d[n][1], 
						//	Rcurr->d[m][0]+1, Rcurr->d[n][0]+1, dist1[ Rcurr->d[m][0] ][ Rcurr->d[n][0] ]);
					}
				}
			}
		} else {
			for(m=0; m<Rcurr->n; m++){
				for(n=0; n<Rcurr->n; n++){
					if( n == m )
						continue;

					if( dist1[ Rcurr->d[m][0] ][ Rcurr->d[n][0] ] <= cut ){
						if( qnames && qnames[Rcurr->d[m][0]] ){
							fprintf(tmpcf, "\n\t %-10s", qnames[Rcurr->d[m][0]]);
						} else {
							fprintf(tmpcf, "\n\t %-10d", Rcurr->d[m][0]+1);
						}

						if( qnames && qnames[Rcurr->d[n][0]] ){
							fprintf(tmpcf, " -- %-10s ", qnames[Rcurr->d[n][0]]);
						} else {
							fprintf(tmpcf, " -- %-10d ", Rcurr->d[n][0]+1);
						}

						if( dist2[ Rcurr->d[m][1] ][ Rcurr->d[n][1] ] > 1 ){
							spath_count=0;
							output_all_pairs_shortest_path(path2, Rcurr->d[m][1], Rcurr->d[n][1]);

							if( nnames && nnames[ spath[0] ] ){
								//fprintf(tmpcf, "  %1s[%d, %d]", nnames[ spath[0] ], spath[0], Rcurr->d[m][1]);
								fprintf(tmpcf, "  %1s", nnames[ spath[0] ]);
							} else {
								fprintf(tmpcf, "  %1d",  spath[0] +1);
							}
							
							for(k=1; k<spath_count; k++){
								if( nnames && nnames[ spath[k] ] ){
									//fprintf(tmpcf, " -- %1s [%d] ", nnames[ spath[k] ], spath[k] );
									fprintf(tmpcf, " -- %1s", nnames[ spath[k] ]);
								} else {
									fprintf(tmpcf, " -- %1d",  spath[k] +1);
								}
							}
							//fprintf(tmpcf, " <%d, %d> ",Rcurr->d[n][1],dist2[ Rcurr->d[m][1] ][ Rcurr->d[n][1] ]); 

						} else {
							if( nnames && nnames[ Rcurr->d[m][1] ] ){
								fprintf(tmpcf, "  %1s", nnames[ Rcurr->d[m][1] ]);
							} else {
								fprintf(tmpcf, "  %1d",  Rcurr->d[m][1] +1);
							}

							if( nnames && nnames[ Rcurr->d[n][1] ] ){
								fprintf(tmpcf, " -- %1s", nnames[ Rcurr->d[n][1] ]);
							} else {
								fprintf(tmpcf, " -- %1d",  Rcurr->d[n][1] +1);
							}						
						}

						//fprintf(tmpcf, "\n\t%5d -- %5d (%2d -- %2d) (%d)", Rcurr->d[m][1], Rcurr->d[n][1], 
						//	Rcurr->d[m][0]+1, Rcurr->d[n][0]+1, dist1[ Rcurr->d[m][0] ][ Rcurr->d[n][0] ]);
					}
				}
			}
		}

		i++;
		Rcurr = Rcurr->next;
	}

	fprintf(tmpcf, "\n");
	fclose(tmpcf);
	return;
}

//use branch & bound to find top cliques
void clique_finding_ng_recurr_tops(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	                               vnode **g, int *gsize, int cut, int n, int l, int dcut){
	int i,j, m, k,x,y,z,  c=0;
	vnode * tmpv;

	for(i=l+1; i< (((n-curr_max_l+1)>=0)?dn1:(dn1+n-curr_max_l+1)); i++){ //pruning
		if( gsize[i] <= 0 )
			continue;

		for(x=0; x<n; x++){
			if( (dist1[ tmpc[x][0] ][ g[i][0].qid ] <= cut) )					
				break;
		}

		if( x < n ){ //there already is a neighbor
			tmpv = & g[ tmpc[x][3] ][ tmpc[x][2] ];
			for(y=0; y<(tmpv->rank)[i]; y++){
				z = ((tmpv->edges)[i][y])->netid; 
				//if(l == 0 )
				//	fprintf(stdout, "\ni=%d y=%d n=%d (%d, %d) con[i]=%d curr_max_l=%d", i, y, n, conds[i][0], z, (tmpv->rank)[i], curr_max_l), fflush(NULL);
				
				if( n + 1 + ((tmpv->edges)[i][y])->ls < curr_max_l)
					continue;

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && 
						(dist2[ tmpc[k][1] ][ z ] > dcut))
						break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = z;
					tmpc[n][2] = ((tmpv->edges)[i][y])->y;
					tmpc[n][3] = i;

					clique_finding_ng_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, n+1, i, dcut);
				}
			}
		} else {
			for(j=0; j<gsize[i]; j++){
				//if(l == 0 )
				//	fprintf(stdout, "\nXXXi=%d j=%d n=%d (%d, %d) con[i]=%d curr_max=%d", i, j, n, conds[i][0], conds[i][j], con[i], curr_max), fflush(NULL);
				
				if( n + 1 + g[i][j].ls < curr_max_l)
					continue;

				//if( curr_max == 0 )
				//	fprintf(stdout, "\n**i=%d j=%d n=%d (%d, %d) con[i]=%d curr_max=%d", i, j, n, conds[i][0], conds[i][j], con[i], curr_max), fflush(NULL);

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && 
						(dist2[ tmpc[k][1] ][ g[i][j].netid ] > dcut) )					
							break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = g[i][j].netid;
					tmpc[n][2] = j;
					tmpc[n][3] = i;

					clique_finding_ng_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, n+1, i, dcut);
				}
			}
		}
	}

	if( !c ){//save maximal clique		

		if( n == dn1 ){
			max_count ++;
			if( max_count % 1000000 == 0 )
					fprintf(stdout, "\ntop%ld  X e6", max_count/1000000), fflush(NULL);
        }

//*
		if( n > curr_max_l ){
			Rcurr = Rhead;
			while( Rcurr->n >= n )
				Rcurr = Rcurr->next;		
			
			if( Rcurr->prev == NULL ){ //head
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev = Rtemp;
				Rtemp->prev = NULL;
				Rtemp->next = Rcurr;
				Rhead = Rtemp;			
				
				Rtemp->n = n;
				//Rtemp->s = 0;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				if(Rtail->n > curr_max_l)
					curr_max_l = Rtail->n;
			} else if ( Rcurr->next == NULL ){ //tail
				Rtail->n = n;
				//Rtemp->s = 0;
				for( m=0; m<n; m++){
					Rtail->d[m][0] = tmpc[m][0];
					Rtail->d[m][1] = tmpc[m][1];
				}

				curr_max_l = n;
			}else {
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev->next = Rtemp;
				Rtemp->next = Rcurr;
				Rtemp->prev = Rcurr->prev;
				Rcurr->prev = Rtemp;

				Rtemp->n = n;
				//Rtemp->s = 0;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				if(Rtail->n > curr_max_l)
					curr_max_l = Rtail->n;
			}
		} 
//*/
	}

	return;
}

//create ng -- correspondence nodes as vertices
void ng_builder(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode ***g, int **gsize, int cut, int dcut){
	int i, j, m, k, n, x, y;
	vnode **h;
	int * hsize;
	double max, t;

	//allocate space for ng
	if( !(hsize = calloc(dn1, sizeof(int))) )
		fprintf(stdout, "out of memory for hsize in ng_builder\n"),fflush(NULL),exit(0);
	*gsize = hsize;

	if( !(h = (vnode **)calloc(dn1, sizeof(vnode *))) )
		fprintf(stdout, "out of memory for h in ng_builder\n"),fflush(NULL),exit(0);
	*g = h;

	for(i=0;i<dn1;i++){
		hsize[i] = con[i] - 1;

		if( hsize[i] > 0 ){
			if( !(h[i] = (vnode *)calloc(hsize[i], sizeof(vnode))) )
				fprintf(stdout, "out of memory for h[] in ng_builder\n"),fflush(NULL),exit(0);
		} else {
			h[i] = NULL;
		}
	}

	//populate h
	for(i=0;i<dn1;i++){
		for(j=0; j<hsize[i]; j++){
			h[i][j].netid = conds[i][j+1];
			h[i][j].qid   = conds[i][0];
			h[i][j].no    = conds[i][0];
			h[i][j].y     = j;

			for(m=0; m<cn; m++){
				if( conds[i][0] == co[m].v1 && conds[i][j+1] == co[m].v2 ){
					h[i][j].s = co[m].lgsim; 
					break;
				} 
			}
			if( m == cn )
				fprintf(stdout, "m==cn in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].rank  =  calloc(dn1, sizeof(int))) )
				fprintf(stdout, "out of memory for h[][].rank in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].edges =  (vnode ***)calloc(dn1, sizeof(vnode **))) )
				fprintf(stdout, "out of memory for h[][].edge in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].ewts  =  (double **)calloc(dn1, sizeof(double *))) )
				fprintf(stdout, "out of memory for h[][].ewts in ng_builder\n"),fflush(NULL),exit(0);

			//rank & edges
			for(k=0; k<dn1; k++){
			//for(k=i+1; k<dn1; k++){
				(h[i][j].rank)[k] = n = 0;
				(h[i][j].edges)[k] = NULL;

				if(conds[k][0] != i){
					if( dist1[ conds[i][0] ][ conds[k][0] ] <= cut ){
						for(m=1; m<con[k]; m++)
							if( dist2[ conds[i][j+1] ][ conds[k][m] ] <= dcut )
								n++;

						(h[i][j].rank)[k] = n;

						if( n ){
							if( !((h[i][j].edges)[k] =  (vnode **)calloc(n, sizeof(vnode *))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							if( !((h[i][j].ewts)[k]  =  (double *)calloc(n, sizeof(double))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							n=0;
							for(m=1; m<con[k]; m++){
								if( dist2[ conds[i][j+1] ][ conds[k][m] ] <= dcut ){
									(h[i][j].edges)[k][n] = & (h[ conds[k][0] ][ m - 1 ]);
									//edge weight
									(h[i][j].ewts)[k][n]  = (dist2[ conds[i][j+1] ][ conds[k][m] ] - 1) * edge_gap_penalty;
									n++;
								}
							}
						}
					} else {
						(h[i][j].rank)[k] = con[k]-1;
					}
				}
			}

			h[i][j].ord   = 0;
			h[i][j].ls    = 0;
			for(k=i+1; k<dn1; k++){
				if( dist1[ conds[i][0] ][ conds[k][0] ] <= cut ){
					if( (h[i][j].rank)[k] > 0 ){
						(h[i][j].ls) ++;

						h[i][j].ord += (h[i][j].rank)[k];
					}
				} else {
					(h[i][j].ls) ++;

					h[i][j].ord += (h[i][j].rank)[k];
				}
			}

			h[i][j].mps   = 0;
		}
	}

	//mps
	for(i=0;i<dn1;i++){
		if( hsize[i] > 0 ){
			for(j=0; j<hsize[i]; j++){
				h[i][j].mps = h[i][j].s;

				for(k=i+1; k<dn1; k++){
					if( hsize[k] > 0 ){
						max = -1e200;
						x   = 0;

						if( dist1[ h[i][0].qid ][ h[k][0].qid ] <= cut ){
							y=0;
							for( m=0; m<hsize[k]; m++ ){
								//if( dist2[ h[i][j].netid ][ h[k][m].netid ] <= dcut && h[k][m].s > max ){
								if( dist2[ h[i][j].netid ][ h[k][m].netid ] <= dcut ){
									t = h[k][m].s + h[i][j].ewts[k][y];
									if( t > max ){
										max = t;
										x = 1;
									}
									y++;
								}
							}
						} else {
							for( m=0; m<hsize[k]; m++ ){
								if( h[k][m].s > max ){
									max = h[k][m].s;
									x = 1; 
								}
							}
						}

						if( x == 1 )
							h[i][j].mps += max;						
					}
				}
			}
		}
	}
	
	fprintf(stdout, "\n\n"), fflush(NULL);
	system("date");
	fprintf(stdout, "\nNG built!\n"), fflush(NULL);
	return;
}


//find cliques in ng -- correspondence nodes as vertices
void clique_finding_ng(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames){
	int i,j;

	Rhead = Rtail = NULL;
	max_count = 0;

	if( if_maximum_clique_only == 1 ){
		curr_max_l = dn1;		
	} else {
		curr_max_l = 0;
		curr_max_w = 0;

		for(j=0; j<top_num; j++){
			if( !Rhead ){
				if( !(Rhead = Rtail = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rhead\n"), fflush(NULL),exit(0);

				Rtail->s = 0;
				Rtail->n = 0; 
				Rtail->prev = NULL;
				Rtail->next = NULL;
				if( !(Rtail->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rtail->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}
			} else {
				if( !(Rcurr = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rcurr\n"), fflush(NULL),exit(0);

				Rcurr->s = 0;
				Rcurr->n = 0; 
				Rcurr->next = NULL;
				if( !(Rcurr->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rcurr->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}

				Rcurr->prev = Rtail;
				Rtail->next = Rcurr;

				Rtail = Rcurr;
			}
		}
	}

	for(i=0; i<=dn1-curr_max_l; i++){
		if( gsize[i] > 0 ){
			for(j=0; j<gsize[i]; j++){
				if( g[i][j].ls +1 >= curr_max_l){
					tmpc[0][0] = g[i][j].qid;
					tmpc[0][1] = g[i][j].netid;
					tmpc[0][2] = j;
					tmpc[0][3] = i;

					fprintf(stdout, "\ni=%d j=%d (%d, %d)  %d", i, j, g[i][j].qid, g[i][j].netid, gsize[i]), fflush(NULL);

					if( if_maximum_clique_only == 1 ){
						//clique_finding_ng_recurr_max(dist1, dn1, dist2, dn2, conds, con, cut, 1, i, dcut);
					} else {
						clique_finding_ng_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, 1, i, dcut);
					}
				}				
			}
		}
	}

	fprintf(stdout, "\ntop%ld", max_count);

	if( if_maximum_clique_only == 0 )
		//output_top_cliques_ng(co, cn, dist1, dn1, dist2, dn2, cut, dcut, qnames, nnames);
	return;
}

//WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
//use branch & bound to find top cliques
void wt_clique_finding_ng_recurr_tops(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	                               vnode **g, int *gsize, int cut, int n, int l, int dcut){
	int i,j, m, k,x,y,z,p, c=0;
	vnode * tmpv, *tmpn;
	double wt;

	for(i=l+1; i< dn1; i++){ //pruning
		if( gsize[i] <= 0 )
			continue;

		for(x=0; x<n; x++){
			if( (dist1[ tmpc[x][0] ][ g[i][0].qid ] <= cut) )					
				break;
		}

		if( x < n ){ //there already is a neighbor
			tmpv = & g[ tmpc[x][3] ][ tmpc[x][2] ];
			for(y=0; y<(tmpv->rank)[i]; y++){
				tmpn = (tmpv->edges)[i][y];
				z = tmpn->netid; 
				//if(l == 0 )
				//	fprintf(stdout, "\ni=%d y=%d n=%d (%d, %d) con[i]=%d curr_max_l=%d", i, y, n, conds[i][0], z, (tmpv->rank)[i], curr_max_l), fflush(NULL);
				
				//if( n + 1 + ((tmpv->edges)[i][y])->ls < curr_max_l)
				//if( tmpcwt[n-1] + tmpn->mps < curr_max_w)
				if( tmpcwt[n-1] + tmpn->mps + (i-l-1)*mnode_penalty < curr_max_w)
					continue;

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && 
						(dist2[ tmpc[k][1] ][ z ] > dcut))
						break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = z;
					tmpc[n][2] = tmpn->y;
					tmpc[n][3] = i;

					tmpcwt[n]  = tmpcwt[n-1] + tmpn->s;

					//add miss node penalty
					tmpcwt[n] += (i-l-1)*mnode_penalty; 

					//add edge weights
					for(p=0; p<n; p++){
						if( (dist1[ tmpc[p][0] ][ tmpc[n][0] ] <= cut) ){
							tmpcwt[n] += (dist2[ tmpc[p][1] ][ z ] - 1)* edge_gap_penalty;
						}
					}

					wt_clique_finding_ng_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, n+1, i, dcut);
				}
			}
		} else {
			for(j=0; j<gsize[i]; j++){
				//if(l == 0 )
				//	fprintf(stdout, "\nXXXi=%d j=%d n=%d (%d, %d) con[i]=%d curr_max=%d", i, j, n, conds[i][0], conds[i][j], con[i], curr_max), fflush(NULL);
				
				//if( n + 1 + g[i][j].ls < curr_max_l)
				if( tmpcwt[n-1] + g[i][j].mps < curr_max_w)
					continue;

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && 
						(dist2[ tmpc[k][1] ][ g[i][j].netid ] > dcut) )					
							break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = g[i][j].netid;
					tmpc[n][2] = j;
					tmpc[n][3] = i;

					tmpcwt[n]  = tmpcwt[n-1] + g[i][j].s;

					wt_clique_finding_ng_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, n+1, i, dcut);
				}
			}
		}
	}

	if( !c ){//save maximal clique		

		if( n == dn1 ){
			max_count ++;
			if( max_count % 1000000 == 0 )
					fprintf(stdout, "\ntop%ld  X e6", max_count/1000000), fflush(NULL);
        }

		//wt = tmpcwt[n-1];
		//add miss node penalty 
		wt = tmpcwt[n-1] + (dn1-n)*mnode_penalty;
		
//*
		//if( n > curr_max_l ){
		if( wt > curr_max_w ){
			Rcurr = Rhead;
			//while( Rcurr->n >= n )
			while( Rcurr->s >= wt )
				Rcurr = Rcurr->next;		
			
			if( Rcurr->prev == NULL ){ //head
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev = Rtemp;
				Rtemp->prev = NULL;
				Rtemp->next = Rcurr;
				Rhead = Rtemp;			
				
				Rtemp->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				//if(Rtail->n > curr_max_l)
				//	curr_max_l = Rtail->n;
				if(Rtail->s > curr_max_w)
					curr_max_w = Rtail->s;
			} else if ( Rcurr->next == NULL ){ //tail
				Rtail->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtail->d[m][0] = tmpc[m][0];
					Rtail->d[m][1] = tmpc[m][1];
				}

				//curr_max_l = n;
				curr_max_w = wt;
			}else {
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev->next = Rtemp;
				Rtemp->next = Rcurr;
				Rtemp->prev = Rcurr->prev;
				Rcurr->prev = Rtemp;

				Rtemp->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				//if(Rtail->n > curr_max_l)
				//	curr_max_l = Rtail->n;
				if(Rtail->s > curr_max_w)
					curr_max_w = Rtail->s;
			}
		} 
//*/
	}

	return;
}


//find cliques with top weights in ng -- correspondence nodes as vertices
void wt_clique_finding_ng(int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, int ** conds, int * con, 
	  pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, char *outfname){
	int i,j;

	Rhead = Rtail = NULL;
	max_count = 0;

	if( if_maximum_clique_only == 1 ){
		curr_max_l = dn1;		
	} else {
		curr_max_l = 0;
		curr_max_w = -1.0e200;

		for(j=0; j<top_num; j++){
			if( !Rhead ){
				if( !(Rhead = Rtail = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rhead\n"), fflush(NULL),exit(0);

				Rtail->s = -1.0e200;
				Rtail->n = 0; 
				Rtail->prev = NULL;
				Rtail->next = NULL;
				if( !(Rtail->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rtail->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}
			} else {
				if( !(Rcurr = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rcurr\n"), fflush(NULL),exit(0);

				Rcurr->s = 0;
				Rcurr->n = 0; 
				Rcurr->next = NULL;
				if( !(Rcurr->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rcurr->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}

				Rcurr->prev = Rtail;
				Rtail->next = Rcurr;

				Rtail = Rcurr;
			}
		}
	}

	//for(i=0; i<=dn1-curr_max_l; i++){
	for(i=0; i<dn1; i++){
		if( gsize[i] > 0 ){
			for(j=0; j<gsize[i]; j++){
				fprintf(stdout, "\ni=%d j=%d %e %e %e", i, j, g[i][j].mps + i*mnode_penalty, g[i][j].s, curr_max_w);

				//if( g[i][j].ls +1 >= curr_max_l){
				if( g[i][j].mps + i*mnode_penalty > curr_max_w ){
					tmpc[0][0] = g[i][j].qid;
					tmpc[0][1] = g[i][j].netid;
					tmpc[0][2] = j;
					tmpc[0][3] = i;

					//tmpcwt[0] = g[i][j].s;
					tmpcwt[0] = g[i][j].s + i*mnode_penalty;

					fprintf(stdout, "\ni=%d j=%d (%d, %d)  %d", i, j, g[i][j].qid, g[i][j].netid, gsize[i]), fflush(NULL);

					if( if_maximum_clique_only == 1 ){
						//clique_finding_ng_recurr_max(dist1, dn1, dist2, dn2, conds, con, cut, 1, i, dcut);
					} else {
						wt_clique_finding_ng_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, 1, i, dcut);
					}
				}				
			}
		}
	}

	fprintf(stdout, "\ntop%ld", max_count);

	if( if_maximum_clique_only == 0 )
		output_top_cliques_ng(co, cn, dist1, dn1, dist2, path2, dn2, cut, dcut, qnames, nnames, outfname);

	return;
}


//DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD
//create directed ng -- correspondence nodes as vertices
void ng_builder_direct(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode ***g, int **gsize, int cut, int dcut){
	int i, j, m, k, n, x, y;
	vnode **h;
	int * hsize;
	double max, t;

	//allocate space for ng
	if( !(hsize = calloc(dn1, sizeof(int))) )
		fprintf(stdout, "out of memory for hsize in ng_builder\n"),fflush(NULL),exit(0);
	*gsize = hsize;

	if( !(h = (vnode **)calloc(dn1, sizeof(vnode *))) )
		fprintf(stdout, "out of memory for h in ng_builder\n"),fflush(NULL),exit(0);
	*g = h;

	for(i=0;i<dn1;i++){
		hsize[i] = con[i] - 1;

		if( hsize[i] > 0 ){
			if( !(h[i] = (vnode *)calloc(hsize[i], sizeof(vnode))) )
				fprintf(stdout, "out of memory for h[] in ng_builder\n"),fflush(NULL),exit(0);
		} else {
			h[i] = NULL;
		}
	}

	//populate h
	for(i=0;i<dn1;i++){
		for(j=0; j<hsize[i]; j++){
			h[i][j].netid = conds[i][j+1];
			h[i][j].qid   = conds[i][0];
			h[i][j].no    = conds[i][0];
			h[i][j].y     = j;

			for(m=0; m<cn; m++){
				if( conds[i][0] == co[m].v1 && conds[i][j+1] == co[m].v2 ){
					h[i][j].s = co[m].lgsim; 
					break;
				} 
			}
			if( m == cn )
				fprintf(stdout, "m==cn in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].rank  =  calloc(dn1, sizeof(int))) )
				fprintf(stdout, "out of memory for h[][].rank in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].edges =  (vnode ***)calloc(dn1, sizeof(vnode **))) )
				fprintf(stdout, "out of memory for h[][].edge in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].ewts  =  (double **)calloc(dn1, sizeof(double *))) )
				fprintf(stdout, "out of memory for h[][].ewts in ng_builder\n"),fflush(NULL),exit(0);

			//rank & edges			
			for(k=0; k<dn1; k++){
			//for(k=i+1; k<dn1; k++){
				(h[i][j].rank)[k] = n = 0;
				(h[i][j].edges)[k] = NULL;

				//if(conds[k][0] != i){
					if( dist1[ conds[i][0] ][ conds[k][0] ] <= cut ){
						for(m=1; m<con[k]; m++)
							if( dist2[ conds[i][j+1] ][ conds[k][m] ] <= dcut )
								n++;

						(h[i][j].rank)[k] = n;

						if( n ){
							if( !((h[i][j].edges)[k] =  (vnode **)calloc(n, sizeof(vnode *))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							if( !((h[i][j].ewts)[k]  =  (double *)calloc(n, sizeof(double))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							n=0;
							for(m=1; m<con[k]; m++){
								if( dist2[ conds[i][j+1] ][ conds[k][m] ] <= dcut ){
									(h[i][j].edges)[k][n] = & (h[ conds[k][0] ][ m - 1 ]);
									//edge weight
									(h[i][j].ewts)[k][n]  = (dist2[ conds[i][j+1] ][ conds[k][m] ] - 1) * edge_gap_penalty;
									n++;
								}
							}
						}
					} else if( dist1[ conds[k][0] ][ conds[i][0] ] <= cut ){
						for(m=1; m<con[k]; m++)
							if( dist2[ conds[k][m] ][ conds[i][j+1] ] <= dcut )
								n++;

						(h[i][j].rank)[k] = n;

						if( n ){
							if( !((h[i][j].edges)[k] =  (vnode **)calloc(n, sizeof(vnode *))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							if( !((h[i][j].ewts)[k]  =  (double *)calloc(n, sizeof(double))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							n=0;
							for(m=1; m<con[k]; m++){
								if( dist2[ conds[k][m] ][ conds[i][j+1] ] <= dcut ){
									(h[i][j].edges)[k][n] = & (h[ conds[k][0] ][ m - 1 ]);
									//edge weight
									(h[i][j].ewts)[k][n]  = (dist2[ conds[k][m] ][ conds[i][j+1] ] - 1) * edge_gap_penalty;
									n++;
								}
							}
						}
					} else {
						(h[i][j].rank)[k] = con[k]-1;
					}
				//}
			}

			h[i][j].ord   = 0;
			h[i][j].ls    = 0;
			for(k=i; k<dn1; k++){
				if( dist1[ conds[i][0] ][ conds[k][0] ] <= cut ){
					if( (h[i][j].rank)[k] > 0 ){
						(h[i][j].ls) ++;

						h[i][j].ord += (h[i][j].rank)[k];
					}
				} else if( dist1[ conds[k][0] ][ conds[i][0] ] <= cut ){
					if( (h[i][j].rank)[k] > 0 ){
						(h[i][j].ls) ++;

						h[i][j].ord += (h[i][j].rank)[k];
					}
				} else {
					(h[i][j].ls) ++;

					h[i][j].ord += (h[i][j].rank)[k];
				}
			}

			h[i][j].mps   = 0;
		}
	}

	//mps
	for(i=0;i<dn1;i++){
		if( hsize[i] > 0 ){
			for(j=0; j<hsize[i]; j++){
				h[i][j].mps = h[i][j].s;

				for(k=i; k<dn1; k++){
					if( hsize[k] > 0 ){
						max = -1e200;
						x   = 0;

						if( dist1[ h[i][0].qid ][ h[k][0].qid ] <= cut ){
							y=0;
							for( m=0; m<hsize[k]; m++ ){
								//if( dist2[ h[i][j].netid ][ h[k][m].netid ] <= dcut && h[k][m].s > max ){
								if( dist2[ h[i][j].netid ][ h[k][m].netid ] <= dcut ){
									t = h[k][m].s + h[i][j].ewts[k][y];
									if( t > max ){
										max = t;
										x = 1;
									}
									y++;
								}
							}
						} else if( dist1[ h[k][0].qid ][ h[i][0].qid ] <= cut ){
							y=0;
							for( m=0; m<hsize[k]; m++ ){
								//if( dist2[ h[i][j].netid ][ h[k][m].netid ] <= dcut && h[k][m].s > max ){
								if( dist2[ h[k][m].netid ][ h[i][j].netid ] <= dcut ){
									t = h[k][m].s + h[i][j].ewts[k][y];
									if( t > max ){
										max = t;
										x = 1;
									}
									y++;
								}
							}
						} else {
							for( m=0; m<hsize[k]; m++ ){
								if( h[k][m].s > max ){
									max = h[k][m].s;
									x = 1; 
								}
							}
						}

						if( x == 1 )
							h[i][j].mps += max;						
					}
				}
			}
		}
	}
	
	fprintf(stdout, "\n\n"), fflush(NULL);
	system("date");
	fprintf(stdout, "\nNG built!\n"), fflush(NULL);
	return;
}


//use branch & bound to find top cliques
void wt_clique_finding_ng_recurr_tops_direct(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	                               vnode **g, int *gsize, int cut, int n, int l, int dcut){
	int i,j, m, k,x,y,z,p, c=0;
	vnode * tmpv, *tmpn;
	double wt;

	//for(i=l+1; i< (((n-curr_max_l+1)>=0)?dn1:(dn1+n-curr_max_l+1)); i++){ //pruning
	for(i=l+1; i< dn1; i++){ //pruning
		if( gsize[i] <= 0 )
			continue;

		for(x=0; x<n; x++){
			if( (dist1[ tmpc[x][0] ][ g[i][0].qid ] <= cut) || (dist1[ g[i][0].qid ][ tmpc[x][0] ] <= cut) )					
				break;
		}

		if( x < n ){ //there already is a neighbor
			tmpv = & g[ tmpc[x][3] ][ tmpc[x][2] ];
			for(y=0; y<(tmpv->rank)[i]; y++){
				tmpn = (tmpv->edges)[i][y];
				z = tmpn->netid; 
				//if(l == 0 )
				//	fprintf(stdout, "\ni=%d y=%d n=%d (%d, %d) con[i]=%d curr_max_l=%d", i, y, n, conds[i][0], z, (tmpv->rank)[i], curr_max_l), fflush(NULL);
				
				//if( n + 1 + ((tmpv->edges)[i][y])->ls < curr_max_l)
				//if( tmpcwt[n-1] + tmpn->mps < curr_max_w)
				if( tmpcwt[n-1] + tmpn->mps + (i-l-1)*mnode_penalty < curr_max_w)
					continue;

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && 
						(dist2[ tmpc[k][1] ][ z ] > dcut) )
						break;

					if( (dist1[ g[i][0].qid ][ tmpc[k][0] ] <= cut) && 
						(dist2[ z ][ tmpc[k][1] ] > dcut) )
						break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = z;
					tmpc[n][2] = tmpn->y;
					tmpc[n][3] = i;

					tmpcwt[n]  = tmpcwt[n-1] + tmpn->s;

					//add miss node penalty
					tmpcwt[n] += (i-l-1)*mnode_penalty; 

					//add edge weights
					for(p=0; p<n; p++){
						if( (dist1[ tmpc[p][0] ][ tmpc[n][0] ] <= cut) ){
							tmpcwt[n] += (dist2[ tmpc[p][1] ][ z ] - 1)* edge_gap_penalty;
						} else if ( (dist1[ tmpc[n][0] ][ tmpc[p][0] ] <= cut) ){
							tmpcwt[n] += (dist2[ z ][ tmpc[p][1] ] - 1)* edge_gap_penalty;
						}
					}

					wt_clique_finding_ng_recurr_tops_direct(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, n+1, i, dcut);
				}
			}
		} else {
			for(j=0; j<gsize[i]; j++){
				//if(l == 0 )
				//	fprintf(stdout, "\nXXXi=%d j=%d n=%d (%d, %d) con[i]=%d curr_max=%d", i, j, n, conds[i][0], conds[i][j], con[i], curr_max), fflush(NULL);
				
				//if( n + 1 + g[i][j].ls < curr_max_l)
				if( tmpcwt[n-1] + g[i][j].mps < curr_max_w)
					continue;

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && 
						(dist2[ tmpc[k][1] ][ g[i][j].netid ] > dcut) )					
							break;

					if( (dist1[ g[i][0].qid ][ tmpc[k][0] ] <= cut) && 
						(dist2[ g[i][j].netid ][ tmpc[k][1] ] > dcut) )					
							break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = g[i][j].netid;
					tmpc[n][2] = j;
					tmpc[n][3] = i;

					tmpcwt[n]  = tmpcwt[n-1] + g[i][j].s;

					wt_clique_finding_ng_recurr_tops_direct(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, n+1, i, dcut);
				}
			}
		}
	}

	if( !c ){//save maximal clique		

		if( n == dn1 ){
			max_count ++;
			if( max_count % 1000000 == 0 )
					fprintf(stdout, "\ntop%ld  X e6", max_count/1000000), fflush(NULL);
        }

		//wt = tmpcwt[n-1];
		//add miss node penalty 
		wt = tmpcwt[n-1] + (dn1-n)*mnode_penalty;
		
//*
		//if( n > curr_max_l ){
		if( wt > curr_max_w ){
			Rcurr = Rhead;
			//while( Rcurr->n >= n )
			while( Rcurr->s >= wt )
				Rcurr = Rcurr->next;		
			
			if( Rcurr->prev == NULL ){ //head
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev = Rtemp;
				Rtemp->prev = NULL;
				Rtemp->next = Rcurr;
				Rhead = Rtemp;			
				
				Rtemp->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				//if(Rtail->n > curr_max_l)
				//	curr_max_l = Rtail->n;
				if(Rtail->s > curr_max_w)
					curr_max_w = Rtail->s;
			} else if ( Rcurr->next == NULL ){ //tail
				Rtail->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtail->d[m][0] = tmpc[m][0];
					Rtail->d[m][1] = tmpc[m][1];
				}

				//curr_max_l = n;
				curr_max_w = wt;
			}else {
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev->next = Rtemp;
				Rtemp->next = Rcurr;
				Rtemp->prev = Rcurr->prev;
				Rcurr->prev = Rtemp;

				Rtemp->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				//if(Rtail->n > curr_max_l)
				//	curr_max_l = Rtail->n;
				if(Rtail->s > curr_max_w)
					curr_max_w = Rtail->s;
			}
		} 
//*/
	}

	return;
}


//find cliques with top weights in ng -- correspondence nodes as vertices
void wt_clique_finding_ng_direct(int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, int ** conds, int * con, 
	  pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, char *outfname){
	int i,j;

	Rhead = Rtail = NULL;
	max_count = 0;

	if( if_maximum_clique_only == 1 ){
		curr_max_l = dn1;		
	} else {
		curr_max_l = 0;
		curr_max_w = -1.0e200;

		for(j=0; j<top_num; j++){
			if( !Rhead ){
				if( !(Rhead = Rtail = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rhead\n"), fflush(NULL),exit(0);

				Rtail->s = -1.0e200;
				Rtail->n = 0; 
				Rtail->prev = NULL;
				Rtail->next = NULL;
				if( !(Rtail->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rtail->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}
			} else {
				if( !(Rcurr = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rcurr\n"), fflush(NULL),exit(0);

				Rcurr->s = -1.0e200;
				Rcurr->n = 0; 
				Rcurr->next = NULL;
				if( !(Rcurr->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rcurr->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}

				Rcurr->prev = Rtail;
				Rtail->next = Rcurr;

				Rtail = Rcurr;
			}
		}
	}

	//for(i=0; i<=dn1-curr_max_l; i++){
	for(i=0; i<dn1; i++){
		if( gsize[i] > 0 ){
			for(j=0; j<gsize[i]; j++){
				fprintf(stdout, "\ni=%d j=%d %e %e %e", i, j, g[i][j].mps + i*mnode_penalty, g[i][j].s, curr_max_w);

				//if( g[i][j].ls +1 >= curr_max_l){
				if( g[i][j].mps + i*mnode_penalty > curr_max_w ){
					tmpc[0][0] = g[i][j].qid;
					tmpc[0][1] = g[i][j].netid;
					tmpc[0][2] = j;
					tmpc[0][3] = i;

					//tmpcwt[0] = g[i][j].s;
					tmpcwt[0] = g[i][j].s + i*mnode_penalty;

					fprintf(stdout, "\ni=%d j=%d (%d, %d)  %d", i, j, g[i][j].qid, g[i][j].netid, gsize[i]), fflush(NULL);

					if( if_maximum_clique_only == 1 ){
						//clique_finding_ng_recurr_max(dist1, dn1, dist2, dn2, conds, con, cut, 1, i, dcut);
					} else {
						wt_clique_finding_ng_recurr_tops_direct(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, 1, i, dcut);
					}
				}				
			}
		}
	}

	fprintf(stdout, "\ntop%ld", max_count);

	if( if_maximum_clique_only == 0 )
		output_top_cliques_ng(co, cn, dist1, dn1, dist2, path2, dn2, cut, dcut, qnames, nnames, outfname);

	return;
}


//CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC CC
//create linked list version of query
void ll_graph(int ** q, int n, int *** ll, int ** ln){
	int i,j, k;

	if( !( *ln = calloc(n, sizeof(int))) )
		fprintf(stdout, "out of memory for *ln\n"),fflush(NULL),exit(0);
	if( !( *ll = calloc(n, sizeof(int *))) )
		fprintf(stdout, "out of memory for *ll\n"),fflush(NULL),exit(0);

	for(i=0; i<n; i++){
		k = 0;
		for(j=0; j<n; j++)
			if( q[i][j] == 1 || q[j][i] == 1 )
				k++;

		(*ln)[i] = k;

		if( k == 0 )
			continue;

		if( !( (*ll)[i] = calloc(k, sizeof(int))) )
			fprintf(stdout, "out of memory for *ll[]\n"),fflush(NULL),exit(0);

		k = 0;
		for(j=0; j<n; j++)
			if( q[i][j] == 1 || q[j][i] == 1 ){
				(*ll)[i][k] = j;
				k++;
			}
	}

	return;
}

//create a new cnode
cnode * createacnode(){
	cnode * t;
	if( !(t = calloc(1, sizeof(cnode))) )
		fprintf(stdout, "out of memory for tc\n"),fflush(NULL),exit(0);
	t->c    = NULL;
	t->next = NULL;

	return t;
}


//create a new tnode
tnode * createatnode(){
	tnode * t;
	if( !(t = calloc(1, sizeof(tnode))) )
		fprintf(stdout, "out of memory for t\n"),fflush(NULL),exit(0);
	t->qid = -1;
	t->t   =  0;
	t->s   = -1e200;
	t->p   = NULL;
	t->d   = NULL;

	return t;
}

int * ppth, * ppth2, * ppth3, *ppthout;
int  currdepth;
tnode ** ppth3tr; 

//check if current path is in the tree
tnode * search_path_intree(int * pth, tnode *tr, int i, int y){
	cnode * currc;

	currc = tr->d;
	while( currc && currc->c->qid != pth[i]){
		currc = currc->next;		
	}
	
	if( currc == NULL ){
		currdepth = i;
		return tr;
	} else {
		if( i == y ){ //the path is in the tree
			if( currc->c->t == 1 ){
				currdepth = 1;	//labelled, done!			
			} else {
				currc->c->t = 1;
				currdepth = 0;  //unlabelled, need to do something!
			}

			return NULL;
		} else {
			return search_path_intree(pth, currc->c, i+1, y);
		}
	}	
}

//insert array into tree
void insert_array_into_tree(int *pth, tnode *t, int x, int y){
	tnode * newt;
	cnode * currc;
	int i;

	if( t == NULL )
		return;

	for(i=x; i<=y; i++){
		newt = createatnode();
		newt->qid = pth[i];
		newt->p   = t;

		currc = t->d;
		if( !currc ){
			t->d = createacnode();
			t->d->c = newt;

			t = newt;
		} else {
			while( currc->next )
				currc = currc->next;
			currc->next = createacnode();
			currc->next->c = newt;

			t = newt;
		}
	}

	t->t = 1;

	return;
}

//compare two numbers for qsort
int comp_nums(const int * n1, const int * n2){
	if( *n1 < *n2 ) {
		return -1;
	} else if( *n1 == *n2) {
		return  0;
	} else
		return  1;
}


//recursive
void recurr_visit_3(int **qg, int **qll, int *qlln,int qsize, tnode *tr, tnode *nr, int x, int y){
	int i,j, cond=0;
	int c[40], cn=0;
	int d[40], dn=0;
	tnode * t, *v;

	for(i=0; i<qsize; i++){
		if( ppthout[ i ] == 0 ){ 
			if ( qg[x][i] == 1 || qg[i][x] ==1 ){
				c[cn] = i;
				cn++;
			} else {
				d[dn] = i;
				dn++;
			}
		} 
	}

	for(i=0; i<cn; i++){
		ppth[y] = c[i];

		for(j=0; j<=y; j++)
			ppth2[j] = ppth[j];		

		//sort the array
		qsort(ppth2, y+1, sizeof(int), (void *)comp_nums);	

		t = search_path_intree(ppth2, tr, 0, y);

		if( t ){
			ppthout[ c[i] ] = 1; 

			insert_array_into_tree(ppth2, t, currdepth, y);

			//ntree
			v = search_path_intree(ppth, nr, 0, y);
			insert_array_into_tree(ppth, v, currdepth, y);
		
			recurr_visit_3(qg, qll, qlln, qsize, tr, nr, c[i], y+1);

			ppthout[ c[i] ] = 0;
		} else if ( !t && currdepth==0 ) { //currdepth == 0 && t == NULL
			ppthout[ c[i] ] = 1; 

			//ntree
			v = search_path_intree(ppth, nr, 0, y);
			insert_array_into_tree(ppth, v, currdepth, y);
		
			recurr_visit_3(qg, qll, qlln, qsize, tr, nr, c[i], y+1);

			ppthout[ c[i] ] = 0;
		}
	}

	for(i=0; i<dn; i++){
		for(j=0; j<y; j++){
			if( qg[ ppth[j] ][ d[i] ] == 1 || qg[ d[i] ][ ppth[j] ] == 1 )
				break;
		}

		if( j == y )
			continue;

		ppth[y] = d[i];
		for(j=0; j<=y; j++)
			ppth2[j] = ppth[j];

		//sort the array
		qsort(ppth2, y+1, sizeof(int), (void *)comp_nums);	
		
		t = search_path_intree(ppth2, tr, 0, y);

		if( t ){
			ppthout[ d[i] ] = 1; 

			insert_array_into_tree(ppth2, t, currdepth, y);

			//ntree
			v = search_path_intree(ppth, nr, 0, y);
			insert_array_into_tree(ppth, v, currdepth, y);

			//recurr_rest(qg, qll, qlln, qsize, tr, nr, y+1);
			recurr_visit_3(qg, qll, qlln, qsize, tr, nr, d[i], y+1);

			ppthout[ d[i] ] = 0;
		}	else if ( !t && currdepth==0 ) { //currdepth == 0 && t == NULL 
			ppthout[ d[i] ] = 1; 

			//ntree
			v = search_path_intree(ppth, nr, 0, y);
			insert_array_into_tree(ppth, v, currdepth, y);

			//recurr_rest(qg, qll, qlln, qsize, tr, nr, y+1);
			recurr_visit_3(qg, qll, qlln, qsize, tr, nr, d[i], y+1);

			ppthout[ d[i] ] = 0;
		}
	}

	return;
}


//build the query tree
void build_query_tree(int ** qg, int **qll, int *qlln, int qsize, tnode **qtree, tnode **ntree){
	int i, j;
	tnode * tr, *nr, *t;

	//initialize the root
	*qtree = tr = createatnode();
	*ntree = nr = createatnode();
	tr->s  = 0;
	nr->s  = 0;

	if( !(ppth = calloc(qsize, sizeof(int))) )
		fprintf(stdout, "out of memory for ppth\n"),fflush(NULL),exit(0);
	if( !(ppth2 = calloc(qsize, sizeof(int))) )
		fprintf(stdout, "out of memory for ppth2\n"),fflush(NULL),exit(0);
	if( !(ppth3 = calloc(qsize, sizeof(int))) )
		fprintf(stdout, "out of memory for ppth3\n"),fflush(NULL),exit(0);
	if( !(ppth3tr = (tnode **)calloc(qsize, sizeof(tnode *))) )
		fprintf(stdout, "out of memory for ppth3tr\n"),fflush(NULL),exit(0);	
	if( !(ppthout = calloc(qsize, sizeof(int))) )
		fprintf(stdout, "out of memory for ppthout\n"),fflush(NULL),exit(0);

	//starting from every node
	for(i=0; i<qsize; i++){
		for(j=0; j<qsize; ++j)  //UPDATE  recurr_output_tree(tnode * tr, int y)
			ppthout[j] = 0;
		ppth2[0] = ppth[0] = i;		

/*
		{
			FILE *outf;
			if( !(outf = fopen("qtree", "a")) )
				fprintf(stdout, "\nFile qtree can't be opened for appending\n"), fflush(NULL), exit(0);

			fprintf(outf, "======= tr tree =========================\n");
			fclose(outf);
			
			recurr_output_tree(tr, 0);
		}

		

		{
			FILE *outf;
			if( !(outf = fopen("qtree", "a")) )
				fprintf(stdout, "\nFile qtree can't be opened for appending\n"), fflush(NULL), exit(0);

			fprintf(outf, "******* nr tree ***********************\n");
			fclose(outf);
			
			recurr_output_tree(nr, 0);
		}
*/
	
		
		if( t = search_path_intree(ppth2, tr, 0, 0) ){
			ppthout[i] = 1;
			insert_array_into_tree(ppth2, t, 0, 0);
/*
		{
			FILE *outf;
			if( !(outf = fopen("qtree", "a")) )
				fprintf(stdout, "\nFile qtree can't be opened for appending\n"), fflush(NULL), exit(0);

			fprintf(outf, "======= tr tree =========================\n");
			fclose(outf);

			recurr_output_tree(tr, 0);
		}
*/
			//ntree
			insert_array_into_tree(ppth, search_path_intree(ppth, nr, 0, 0), 0, 0);
/*
		{
			FILE *outf;
			if( !(outf = fopen("qtree", "a")) )
				fprintf(stdout, "\nFile qtree can't be opened for appending\n"), fflush(NULL), exit(0);

			fprintf(outf, "******* nr tree ***********************\n");
			fclose(outf);
			recurr_output_tree(nr, 0);		
		}
*/			
			recurr_visit_3(qg, qll, qlln, qsize, tr, nr, i, 1);
			ppthout[i] = 0;
		}		
	}
}


//calculate maximally possible scores in the guide tree
void recurr_score_tree(tnode * tr, int y, int ** dist1, int dn1,int ** dist2, int dn2, vnode **g, int *gsize){
	cnode * c;
	int i, j;
	
	if( y > 1 ){
		for(i=0; i<y-1; i++){
			if( dist1[ppth3[i]][ ppth3[y-1] ] == 1 )
				break;
		}

		if( i < y-1 ){ //there is an edge to y
			tr->s = -1e200;
			for(j=0; j<gsize[ppth3[y-1]]; j++){
				if( g[ppth3[y-1]][j].rank[ppth3[i]] > 0 && g[ppth3[y-1]][j].s > tr->s )
					tr->s = g[ppth3[y-1]][j].s;
			}
		} else { //no edge
			tr->s = -1e200;
			for(j=0; j<gsize[ppth3[y-1]]; j++){
				if( g[ppth3[y-1]][j].s > tr->s )
					tr->s = g[ppth3[y-1]][j].s;
			}
		}
	} else if( y == 1 ){
		tr->s = -1e200;
		for(j=0; j<gsize[ppth3[0]]; j++){
			if( g[ppth3[0]][j].s > tr->s )
				tr->s = g[ppth3[0]][j].s;
		}
	} 

	if( ! tr->d ){	
		tr->x  = 0;
	} else {
		double max=-1e200;

		c = tr->d;
		while( c ){
			ppth3[y] = c->c->qid;
			recurr_score_tree(c->c, y+1, dist1, dn1, dist2, dn2, g, gsize);

			if( c->c->s > max )
				max = c->c->s;

			c = c->next;
		}

		tr->s += max;
		tr->x  = max;
	}
}



//recurr_output output the guide tree
void recurr_output_tree(tnode * tr, int y){
	cnode * c;

	if( ! tr->d ){
		if( y ){
			FILE *outf;
			int i;
			if( !(outf = fopen("qtree", "a")) )
				fprintf(stdout, "\nFile qtree can't be opened for appending\n"), fflush(NULL), exit(0);

			//fprintf(outf, "\n\n==>output tree ");
			for(i=0; i<y; i++){
				fprintf(outf, "%d ", ppth3[i]), fflush(NULL);
			}
			fprintf(outf, "\n"), fflush(NULL);
			for(i=0; i<y; i++){
				fprintf(outf, "%8.3f ", (ppth3tr[i])->x), fflush(NULL);
			}
			fprintf(outf, "\n"), fflush(NULL);

			{
				int i;
				fprintf(outf, "\nppthout: "), fflush(NULL);
				for(i=0; i<14; i++)
					fprintf(outf, " %d", ppthout[i]), fflush(NULL);

				fprintf(outf, "\nppth: "), fflush(NULL);
				for(i=0; i<14; i++)
					fprintf(outf, " %d", ppth[i]), fflush(NULL);

				fprintf(outf, "\nppth2: "), fflush(NULL);
				for(i=0; i<14; i++)
					fprintf(outf, " %d", ppth2[i]), fflush(NULL);
				fprintf(outf, "\n"), fflush(NULL);
			}		

			//fprintf(outf, "<==output tree end!\n\n");
			fclose(outf);
		}
	} else {
		c = tr->d;
		while( c ){
			ppth3[y] = c->c->qid;
			ppth3tr[y] = c->c;
			recurr_output_tree(c->c, y+1);

			c = c->next;
		}
	}
}



//create ng -- correspondence nodes as vertices
void ng_builder_cc(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	             pair * co, int cn, vnode ***g, int **gsize, int cut, int dcut){
	int i, j, m, k, n;
	vnode **h;
	int * hsize;
	//double max, t; int x, y;

	//allocate space for ng
	if( !(hsize = calloc(dn1, sizeof(int))) )
		fprintf(stdout, "out of memory for hsize in ng_builder\n"),fflush(NULL),exit(0);
	*gsize = hsize;

	if( !(h = (vnode **)calloc(dn1, sizeof(vnode *))) )
		fprintf(stdout, "out of memory for h in ng_builder\n"),fflush(NULL),exit(0);
	*g = h;

	for(i=0;i<dn1;i++){
		hsize[i] = con[i] - 1;

		if( hsize[i] > 0 ){
			if( !(h[i] = (vnode *)calloc(hsize[i], sizeof(vnode))) )
				fprintf(stdout, "out of memory for h[] in ng_builder\n"),fflush(NULL),exit(0);
		} else {
			h[i] = NULL;
		}
	}

	//populate h
	for(i=0;i<dn1;i++){
		for(j=0; j<hsize[i]; j++){
			h[i][j].netid = conds[i][j+1];
			h[i][j].qid   = conds[i][0];
			h[i][j].no    = conds[i][0];
			h[i][j].y     = j;

			for(m=0; m<cn; m++){
				if( conds[i][0] == co[m].v1 && conds[i][j+1] == co[m].v2 ){
					h[i][j].s = co[m].lgsim; 
					break;
				} 
			}
			if( m == cn )
				fprintf(stdout, "m==cn in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].rank  =  calloc(dn1, sizeof(int))) )
				fprintf(stdout, "out of memory for h[][].rank in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].edges =  (vnode ***)calloc(dn1, sizeof(vnode **))) )
				fprintf(stdout, "out of memory for h[][].edge in ng_builder\n"),fflush(NULL),exit(0);

			if( !(h[i][j].ewts  =  (double **)calloc(dn1, sizeof(double *))) )
				fprintf(stdout, "out of memory for h[][].ewts in ng_builder\n"),fflush(NULL),exit(0);

			//rank & edges
			for(k=0; k<dn1; k++){
			//for(k=i+1; k<dn1; k++){
				(h[i][j].rank)[k] = n = 0;
				(h[i][j].edges)[k] = NULL;

				if(conds[k][0] != i){
					if( dist1[ conds[i][0] ][ conds[k][0] ] <= cut ){
						for(m=1; m<con[k]; m++)
							if( dist2[ conds[i][j+1] ][ conds[k][m] ] <= dcut )
								n++;

						(h[i][j].rank)[k] = n;

						if( n ){
							if( !((h[i][j].edges)[k] =  (vnode **)calloc(n, sizeof(vnode *))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							if( !((h[i][j].ewts)[k]  =  (double *)calloc(n, sizeof(double))) )
								fprintf(stdout, "out of memory for h[][].edge[] in ng_builder\n"),fflush(NULL),exit(0);

							n=0;
							for(m=1; m<con[k]; m++){
								if( dist2[ conds[i][j+1] ][ conds[k][m] ] <= dcut ){
									(h[i][j].edges)[k][n] = & (h[ conds[k][0] ][ m - 1 ]);
									//edge weight
									(h[i][j].ewts)[k][n]  = (dist2[ conds[i][j+1] ][ conds[k][m] ] - 1) * edge_gap_penalty;
									n++;
								}
							}
						}
					} else {
						(h[i][j].rank)[k] = con[k]-1;
					}
				}
			}

			h[i][j].ord   = 0;
			h[i][j].ls    = 0;
	
			h[i][j].mps   = 0;
		}
	}

	//fprintf(stdout, "\n\n"), fflush(NULL);
	//system("date");
	//fprintf(stdout, "\nNG built!\n"), fflush(NULL);
	return;
}


//use branch & bound to find top cliques
void wt_clique_finding_ng_cc_recurr_tops(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, vnode **g, 
	                   int *gsize, int cut, int n, int dcut, int yyy, pair * co, int cn, char **qnames, char **nnames, int ** path2){
	int i,j, m, k,x,y,z,p, c=0;
	vnode * tmpv, *tmpn;
	double wt;

	i = ppth3[n];
	if( n < yyy && gsize[ i ] > 0 ){
		for(x=0; x<n; x++){
			if( (dist1[ tmpc[x][0] ][ g[i][0].qid ] <= cut) )					
				break;
		}

		if( x < n ){ //there already is a neighbor
			tmpv = & g[ tmpc[x][3] ][ tmpc[x][2] ];
			for(y=0; y<(tmpv->rank)[i]; y++){
				tmpn = (tmpv->edges)[i][y];
				z = tmpn->netid; 
				
				if( tmpcwt[n-1] + tmpn->s + ppth3tr[n]->x <= curr_max_w)
					continue;

				for(k=0; k<n; k++){
					if( ((dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut) && (dist2[ tmpc[k][1] ][ z ] > dcut))
					 || ((dist1[ g[i][0].qid ][ tmpc[k][0] ] <= cut) && (dist2[ z ][ tmpc[k][1] ] > dcut)) )
						break;					
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = z;
					tmpc[n][2] = tmpn->y;
					tmpc[n][3] = i;

					tmpcwt[n]  = tmpcwt[n-1] + tmpn->s;
				
					//add edge weights
					for(p=0; p<n; p++){						
						if( (dist1[ tmpc[p][0] ][ tmpc[n][0] ] <= cut) ){
							tmpcwt[n] += (dist2[ tmpc[p][1] ][ z ] - 1)* edge_gap_penalty;
						} else if ( (dist1[ tmpc[n][0] ][ tmpc[p][0] ] <= cut) ) {
							tmpcwt[n] += (dist2[ z ][ tmpc[p][1] ] - 1)* edge_gap_penalty;
						}
					}

					wt_clique_finding_ng_cc_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, 
						 cut, n+1, dcut, yyy, co, cn, qnames, nnames, path2);
				}
			}
		} else {
			for(j=0; j<gsize[i]; j++){
				//if( tmpcwt[n-1] + g[i][j].mps < curr_max_w)
				if( tmpcwt[n-1] + g[i][j].s + ppth3tr[n]->x <= curr_max_w)
					continue;

				for(k=0; k<n; k++){
					if( (dist1[ tmpc[k][0] ][ g[i][0].qid ] <= cut && dist2[ tmpc[k][1] ][ g[i][j].netid ] > dcut) 
					 || (dist1[ g[i][0].qid ][ tmpc[k][0] ] <= cut && dist2[ g[i][j].netid ][ tmpc[k][1] ] > dcut) )					
							break;
				}

				if( k == n ){
					c = 1;

					tmpc[n][0] = g[i][0].qid;
					tmpc[n][1] = g[i][j].netid;
					tmpc[n][2] = j;
					tmpc[n][3] = i;

					tmpcwt[n]  = tmpcwt[n-1] + g[i][j].s;

					//add edge weights
					for(p=0; p<n; p++){						
						if( (dist1[ tmpc[p][0] ][ tmpc[n][0] ] <= cut) ){
							tmpcwt[n] += (dist2[ tmpc[p][1] ][ tmpc[n][1] ] - 1)* edge_gap_penalty;
						} else if ( (dist1[ tmpc[n][0] ][ tmpc[p][0] ] <= cut) ) {
							tmpcwt[n] += (dist2[ tmpc[n][1] ][ tmpc[p][1] ] - 1)* edge_gap_penalty;
						}
					}

					wt_clique_finding_ng_cc_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, 
						                                cut, n+1, dcut, yyy, co, cn, qnames, nnames, path2);
				}
			}
		}
	}

	if( !c ){//save maximal clique	
		wt = tmpcwt[n-1] + (yyy-n)*mnode_penalty;
	
		if( wt > curr_max_w ){
			if( wt < -3.59 && wt > -3.62 ){
				int xxxxxxxxxxx=0;
				Rcurr = Rhead;
				while( Rcurr ){
					printf("%2d %7.2f\n",++xxxxxxxxxxx, Rcurr->s );
					Rcurr = Rcurr->next;
				}

				xxxxxxxxxxx=0;
				Rcurr = Rhead;
				while( Rcurr && Rcurr->s >= wt ){
					printf("%2d %7.2f\n",++xxxxxxxxxxx, Rcurr->s );
					Rcurr = Rcurr->next;
				}
			}

			Rcurr = Rhead;
			while( Rcurr && Rcurr->s >= wt )
				Rcurr = Rcurr->next;		
			
			if( Rcurr->prev == NULL ){ //head
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rcurr->prev = Rtemp;
				Rtemp->prev = NULL;
				Rtemp->next = Rcurr;
				Rhead = Rtemp;			
				
				Rtemp->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				if(Rtail->s > curr_max_w)
					curr_max_w = Rtail->s;
			} else if ( Rcurr->next == NULL ){ //tail
				Rtail->n = n;
				Rtail->s = wt;
				for( m=0; m<n; m++){
					Rtail->d[m][0] = tmpc[m][0];
					Rtail->d[m][1] = tmpc[m][1];
				}

				curr_max_w = wt;
			}else {
				Rtemp = Rtail;

				Rtail = Rtail->prev; 
				Rtail->next = NULL;

				Rtemp->prev = NULL;
				Rcurr->prev->next = Rtemp;
				Rtemp->next = Rcurr;
				Rtemp->prev = Rcurr->prev;
				Rcurr->prev = Rtemp;

				Rtemp->n = n;
				Rtemp->s = wt;
				for( m=0; m<n; m++){
					Rtemp->d[m][0] = tmpc[m][0];
					Rtemp->d[m][1] = tmpc[m][1];
				}

				if(Rtail->s > curr_max_w)
					curr_max_w = Rtail->s;
			}

			//tmpwwtt = wt;
			//output_top_cliques_ng(co, cn, dist1, dn1, dist2, path2, dn2, cut, dcut, qnames, nnames);
		} 
	}

	return;
}



//find cliques with top weights in ng -- correspondence nodes as vertices
void wt_clique_finding_ng_cc(int ** dist1, int dn1,int ** dist2, int dn2, int ** conds, int * con, 
	  pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, int y, char **qnames, char **nnames, int ** path2){
	int i,j;

	i = ppth3[0];

	if( gsize[i] > 0 ){
		for(j=0; j<gsize[i]; j++){
			//if( g[i][j].mps > curr_max_w ){
			//if( (ppth3tr[0])->x + g[i][j].s + (dn1-y)*mnode_penalty > curr_max_w ){
			if( (ppth3tr[0])->x + g[i][j].s + (dn1-y)*mnode_penalty > curr_max_w ){
				tmpc[0][0] = g[i][j].qid;
				tmpc[0][1] = g[i][j].netid;
				tmpc[0][2] = j;
				tmpc[0][3] = i;

				tmpcwt[0] = g[i][j].s + (dn1-y)*mnode_penalty;
				//tmpcwt[0] = g[i][j].s + i*mnode_penalty;

				wt_clique_finding_ng_cc_recurr_tops(dist1, dn1, dist2, dn2, conds, con, g, gsize, cut, 1, 
					                                dcut, y, co, cn, qnames, nnames, path2);
			}				
		}
	}

	return;
}

//recurr_output output the guide tree
void recurr_cc_tree(tnode * tr, int y, int ** d1, int dn1,int ** d2, int dn2, int ** conds, 
	 int * con, pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, int ** path2){
	cnode * c;

	if( ! tr->d ){
		if( y ){
			wt_clique_finding_ng_cc(d1,dn1,d2, dn2, conds, con, co, cn, g, gsize,cut, dcut, y, qnames, nnames, path2);			
		}
	} else {
		c = tr->d;
		while( c ){
			ppth3[y] = c->c->qid;
			ppth3tr[y] = c->c;
			recurr_cc_tree(c->c, y+1, d1, dn1,d2,dn2,conds,con,co,cn, g, gsize,cut,dcut, qnames, nnames, path2);

			c = c->next;
		}
	}
}


//find connected components with top weights in ng -- correspondence nodes as vertices
void cc_subgraph(tnode * tr, int ** dist1, int dn1,int ** dist2, int ** path2, int dn2, int ** conds, int * con, 
	  pair * co, int cn, vnode **g, int *gsize, int cut, int dcut, char **qnames, char **nnames, char * outfname){
	int i,j;

	Rhead = Rtail = NULL;
	max_count = 0;

	if( if_maximum_clique_only == 1 ){
		curr_max_l = dn1;		
	} else {
		curr_max_l = 0;
		curr_max_w = -1.0e200;

		for(j=0; j<top_num; j++){
			if( !Rhead ){
				if( !(Rhead = Rtail = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rhead\n"), fflush(NULL),exit(0);

				Rtail->s = -1.0e200;
				Rtail->n = 0; 
				Rtail->prev = NULL;
				Rtail->next = NULL;
				if( !(Rtail->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rtail->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}
			} else {
				if( !(Rcurr = calloc(1, sizeof(rnode))) )
					fprintf(stdout, "out of memory for Rcurr\n"), fflush(NULL),exit(0);

				Rcurr->s = -1.0e200;
				Rcurr->n = 0; 
				Rcurr->next = NULL;
				if( !(Rcurr->d = calloc(max_clique, sizeof(int *))) )
					fprintf(stdout, "out of memory for Rtail->d\n"), fflush(NULL),exit(0);
				for(i=0; i<max_clique; i++){
					if( !(Rcurr->d[i] = calloc(2, sizeof(int))) ){
						fprintf(stdout, "out of memory for Rtail->d[]\n"), fflush(NULL),exit(0);
					}
				}

				Rcurr->prev = Rtail;
				Rtail->next = Rcurr;

				Rtail = Rcurr;
			}
		}
	}

	recurr_cc_tree(tr, 0, dist1, dn1,dist2, dn2, conds, con, co, cn,g,gsize,cut, dcut, qnames, nnames, path2);

	if( if_maximum_clique_only == 0 )
		output_top_cliques_ng(co, cn, dist1, dn1, dist2, path2, dn2, cut, dcut, qnames, nnames, outfname);

	return;
}
